C:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
C NAME
C     blocksize -- Determine the optimal block size for distributing
C     a matrix across processes
C
C SYNOPSIS
      Logical Function BlockSize( NProc, Dim1, Dim2, Min1, Min2,
     $   Blk1, Blk2)
      Implicit NONE
      Integer NProc, Dim1, Dim2, Min1, Min2, Blk1, Blk2
C
C ARGUMENTS
C     NProc   Number of processes to distribute over (IN)
C     Dim1    Row dimension of matrix (IN)
C     Dim2    Column dimension of matrix (IN)
C     Min1    Minimum block size in row direction (IN)
C     Min2    Minimum block size in column direction (IN)
C     Blk1    Block size in row direction (OUT)
C     Blk2    Block size in column direction (OUT)
C     
C DESCRIPTION
C     Determine a blocksize for distribution of a 2d array over NPROC
C     processes.  MIN1 and MIN2 give minimum blocksizes.  If they
C     cannot be obeyed, the function returns .FALSE. to indicate an
C     error.  Otherwise it returns .TRUE. with the block sizes in 
C     BLK1 and BLK2.
C
C     This algorithm is shamelessly lifted from RJH's initial
C     implementation of blocking in the GA_Create global array 
C     routine.
C
C     It would be nice if some smarts were added to the piece where
C     the block sizes a bumped up of they are too small so that when
C     possible, block size that are exact divisors of their corresponding
C     dimension are retained.  This will lead to less wasted space.
C
C REVISION
C     $Id: blocksize.f,v 1.1 1993/11/08 19:49:39 gg502 Exp $
C:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
C LOCAL VARIABLES
      Integer NChunk
C
C     Check inputs for sanity
C
      If ( NProc .le. 0
     $   .OR. Dim1 .lt. 0 .OR. Dim2 .lt. 0
     $   .OR. Min1 .lt. 0 .OR. Min2 .lt. 0
     $   .OR. Min1 .gt. Dim1 .OR. Min2 .gt. Dim2) then
         BlockSize = .FALSE.
         Return
      EndIf
C
C     Handle null inputs elegantly
C
      If ( Dim1 * Dim2 .eq. 0) then
         Blk1 = 0
         Blk2 = 0
         BlockSize = .TRUE.
         Return
      EndIf
C
C     Try to figure out an OK chunking ... want:
C        a) at most one chunk per processor
C        b) chunk dimensions should be >= user provided limits
C        c) chunks should be as square as possible (assumes starts square)
C        
C     1) Guess at a blocking
C
      If (Min1.le.1 .and. Min2 .le. 1) Then
         Blk1 = Max(1, dim1/int(dsqrt(dfloat(NProc))))
         Blk2 = Max(1, (dim1*dim2)/(NProc*Blk1))
      ElseIf (Min1.le.1) Then
         Blk1 = Max(1, (dim1*dim2)/(NProc*Min2))
         Blk2 = Min2
      ElseIf (Min2.le.1) Then
         Blk1 = Min1
         Blk2 = Max(1, (dim1*dim2)/(NProc*Min1))
      Else
         Blk1 = Max(1,Min1)
         Blk2 = Max(1,Min2)
      Endif
C
C     2) See how many panels this blocking creates
C
 10   continue
      nchunk = ((dim1-1)/Blk1 + 1)*((dim2-1)/Blk2 + 1)
C
C     3) If there are too many panels, increase the size of the blocks.
C
      If (nchunk .gt. NProc) Then
         If (Blk1.eq.dim1 .and. Blk2.eq.dim2) Then
            BlockSize = .FALSE.
            Return
         EndIf
C
         If (Blk1 .lt. dim1) Blk1 = Blk1 + 1
         If (Blk2 .lt. dim2) Blk2 = Blk2 + 1
C
         Goto 10
      Endif
C
      BlockSize = .TRUE.
      Return
      End
