C:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
C NAME
C     Par1_make_kop -- Form the exchange operators, K_{ij}
C
C SYNOPSIS
      Subroutine Par1_Make_KOp( C, KOp_Hdl, NAO, NOcc, TotInt,
     $   FPMO, FPAO)
      Implicit NONE
      Integer NAO, NOcc, KOp_Hdl
      Double precision TotInt, FPMO, FPAO
      Double Precision C(NAO, NOcc)
C
C ARGUMENTS
C     C       MO coefficients of the occupied orbitals (IN)
C     KOp_Hdl GA handle for exchange operators (OUT)
C     NAO     Number of AO basis functions (IN)
C     NOcc    Number of occupied orbitals (IN)
C     TotInt  Number of AO integrals actually processed (OUT)
C     FPMO    Number of floating point multiply operations done (OUT)
C     FPAO    Number of floating point add operations done (OUT)
C
C DESCRIPTION
C     Form the exchange operators from AO integrals using an O(o^2N^4)
C     algorithm and a single pass through the integrals.
C
C     The resulting operators are NOT symmetric in their AO indices.
C
C MEMORY USAGE
C     Allocates space to generate two-electron integrals on stack (space
C     required given by Mem_TwoInt) plus NAO*NAO MT_DBLs for the "density"
C     matrix used in the transformation.
C
C NOTES
C     At present, no checking is done of sparsity in the integrals or in
C     the "density", although some sparsity would be expected in both
C     for large systems.  Integrals could be screened using Schwartz's
C     inequality.  For the densities it is a little tougher, since
C     there are actually NOcc*(NOcc+1)/2 different densities, more space
C     would have to be devoted to the screening matrices than might be
C     desirable.  Alternatively, one could implement the screening based
C     on the largest value in each shell block across all of the
C     NOcc*(NOcc+1)/2 densities.  This will probably represent a significant
C     reduction in the sparsity, but may still help a little.
C
C REVISION
C     $Id: par1_make_kop.F,v 1.4 1993/11/08 19:49:51 gg502 Exp $
C:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#include "stdio.fh"
#include "integral_api.fh"
#include "numerical_constants.fh"
#include "precision.h"
#ifdef TCGMSG
#include "tcgmsg.h"
#endif /* TCGMSG */
C FUNCTIONS
      Double Precision DDOT, DNRM2
      External DDOT, DNRM2
C
C LOCAL VARIABLES
C
C     For integral evaluation & other dynamically allocated
C     work space.
C
      Integer IEnv, Val_Ind, P_Ind, Q_Ind, R_Ind, S_Ind, MxNInt,
     $    Scr_Ind, LScratch, NShell, D_Hdl, D_Ind, LclK_Hdl, LclK_Ind
      Logical Canonical_Ints, Nonzero_Ints, LResult
C
C     Indices, etc. required for processing integrals
C
      Integer I, J, P, Q, R, S, IJ, N, NInts
      Logical Perm_Sym
      Double precision Fact
C
C     Shell indices and info about basis functions in shell
C
      Integer PShell, QShell, RShell, SShell, SHi
      Integer PStart, QStart, RStart, SStart
      Integer PEnd, QEnd, REnd, SEnd
      Integer PLen, QLen, RLen, SLen
C
C     For determining the local K_{ij}
C
      Integer AOLo, AOHi, IJLo, IJHi, Junk
C
C     Timings & Operation counts
C
      Double precision T_Int, T_Den, T_KOp, T_Upd
      Integer Count
C
C STATEMENT FUNCTIONS
C     These statement functions are intended to hide a little of the
C     addressing complexity involved in locally allocated arrays.
C     These functions return appropriate offsets for use within DBL_MB
C     or INT_MB.
C
      Integer ERI, D, LclK
      Integer Idx1, Idx2, Idx3, Idx4
C
      ERI(  N ) = Val_Ind + N - 1
C
      Idx1( N ) = P_Ind + N - 1
      Idx2( N ) = Q_Ind + N - 1
      Idx3( N ) = R_Ind + N - 1
      Idx4( N ) = S_Ind + N - 1
C
      D(P, Q) = D_Ind + P + NAO*(Q-1) - 1
      LclK(P, Q) = LclK_Ind + P + NAO*(Q-1) - 1
C
C     ***********************
C     * Zero out statistics *
C     ***********************
C
      TotInt = FP_Zero
      FPMO = FP_Zero
      FPAO = FP_Zero
C
      T_Int = FP_Zero
      T_Den = FP_Zero
      T_KOp = FP_Zero
      T_Upd = FP_Zero
      Count = 0
C
C     **********************************
C     * Prepare KOp to accumulate data *
C     **********************************
C
      Call GA_Zero( KOp_Hdl)
C
C     ********************************************
C     * Find out what part of KOp we should make *
C     ********************************************
C
      Call GA_Distribution( KOp_Hdl, NodeID(), AOLo, AOHi, IJLo, IJHi,
     $   Junk)
C
C     If there is no work to do, we can quit early.  Note that
C     all MA allocations occur AFTER this point, so we will skip
C     both the allocation and deallocation.  No matter when we
C     leave this routine, the stack must end up looking as it did
C     on entry.
C
      If (IJLo .eq. 0 .AND. IJHi .eq. -1) then
         Write (LuOut, *) 'Par1_Make_KOp: No work for process ',
     $      NodeID()
         Return
      EndIf
C
C     *******************************************************
C     * Allocate memory for integral routines and "density" *
C     *******************************************************
C
      LResult = Push_TwoInt()
      LResult = Get_Index_TwoInt( IEnv, Val_Ind, P_Ind, Q_Ind, R_Ind,
     $    S_Ind, MxNInt, Scr_Ind, LScratch)
      If ( .NOT. LResult ) Then
         Write (LuErr, *) ' Bad news trying to get indices.'
         STOP
      EndIf
C
      LResult = MA_Push_Get( MT_Dbl, NAO*NAO, 'Density for KOp',
     $   D_Hdl, D_Ind)
      LResult = MA_Push_Get( MT_Dbl, NAO*NAO, 'Local matrix KOp',
     $   LclK_Hdl, LclK_Ind)
C
C     ********************************************
C     * Control massaging of integrals by TwoInt *
C     ********************************************
C     The indices need not be completely canonical.
C     Better to remove zero integrals once than process them many times???
C
      Canonical_Ints = .FALSE.
      Nonzero_Ints = .TRUE.
C
C     *****************************************
C     * Loop unique shells in canonical order *
C     *****************************************
C
      Call GNShell( IEnv, NShell)
      Do PShell = 1, nshell
         Call BFromS( PShell, PStart, PEnd)
         PLen = PEnd - PStart + 1
C
         Do QShell = 1, PShell
            Call BFromS( QShell, QStart, QEnd)
            QLen = QEnd - QStart + 1
C
            Do RShell = 1, PShell
               Call BFromS( RShell, RStart, REnd)
               RLen = REnd - RStart + 1
               SHi = RShell
               if (PShell .eq. RShell) SHi = QShell
C
               Do SShell = 1, SHi
                  Call BFromS( SShell, SStart, SEnd)
                  SLen = SEnd - SStart + 1
C
C                 Generate a batch of integrals
C
                  T_Int = T_Int - TCGTime()
                  Call TwoInt(IEnv, PShell, QShell, RShell, SShell,
     $               DBL_MB( ERI(1)), Int_MB( Idx1(1)),
     $               Int_MB( Idx2(1)), Int_MB( Idx3(1)),
     $               Int_MB( Idx4(1)),
     $               NInts, Dbl_MB(Scr_Ind), LScratch,
     $               Canonical_Ints, Nonzero_Ints)
C
                  TotInt = TotInt + NInts
C
C                 Scale the integrals according to their permutational
C                 symmetry.  This can only happen if the shell indices
C                 match up, so only do the loop if there will be some
C                 scaling done.
C
                  Perm_Sym = .FALSE.
                  If ( PShell .eq. QShell) then
                     Perm_Sym = .TRUE.
                     FPMO = FPMO + PLen
                  EndIf
                  If ( RShell .eq. SShell) then
                     Perm_Sym = .TRUE.
                     FPMO = FPMO + RLen
                  EndIf
                  If ( PShell .eq. RShell) then
                     Perm_Sym = .TRUE.
                     FPMO = FPMO + PLen
                  EndIf
C
                  If (Perm_Sym) then
C
C                    Setting Fact is accounted for above.  This is
C                    the work of applying Fact to the integral.
C
                     FPMO = FPMO + NInts
C
                     Do N = 1, NInts
                        P = Int_MB( Idx1( N ) )
                        Q = Int_MB( Idx2( N ) )
                        R = Int_MB( Idx3( N ) )
                        S = Int_MB( Idx4( N ) )
C
                        Fact = FP_One
                        If ( P .eq. Q) Fact = Fact/2
                        If (R .eq. S) Fact = Fact /2
                        If (P .eq. R .AND. Q .eq. S) Fact = Fact/2
C                    
C                       On the lhs, we must use the full expression
C                       for ERI(N)
C
                        Dbl_MB( ERI(N) ) = Fact * DBL_MB( ERI(N) )
C                    
                     EndDo
                  EndIf
                  T_Int = T_Int + TCGTime()
C                 
C                 Loop over local IJ pairs
C                 
                  Do IJ = IJLo, IJHi
                     Call Detriangularize(IJ, I, J)
C
C                    Form a density-like beast, D_{\mu \nu}^{ij}
C                    from the I and J columns of the MO coefficients.
C                    Note that this matrix is NOT symmetric!
C
C                    Notice that we are presently inside the loop
C                    over shells.  This means that we will construct
C                    each D_{ij} many times.  For each shell quartet,
C                    however, only a few blocks of D are actually
C                    needed for the calculation.  Therefore, we
C                    form only the eight blocks of D which are needed
C                    for this shell quartet.  This is a significant
C                    savings, especially when shells are small!
C
                     FPMO = FPMO + 2 * PLen * RLen
                     FPMO = FPMO + 2 * PLen * SLen
                     FPMO = FPMO + 2 * QLen * RLen
                     FPMO = FPMO + 2 * QLen * SLen
C                    
                     FPAO = FPAO + 2 * PLen * RLen
                     FPAO = FPAO + 2 * PLen * SLen
                     FPAO = FPAO + 2 * QLen * RLen
                     FPAO = FPAO + 2 * QLen * SLen
C
                     T_Den = T_Den - TCGTime()
                     Call DGEMM('N','T', PLen, RLen, 1,
     $                  FP_One, C(PStart, I), NAO, C(RStart, J), NAO,
     $                  FP_Zero, Dbl_MB( D(PStart, RStart)), NAO)
C
                     Call DGEMM('N','T', PLen, SLen, 1,
     $                  FP_One, C(PStart, I), NAO, C(SStart, J), NAO,
     $                  FP_Zero, Dbl_MB( D(PStart, SStart)), NAO)
C
                     Call DGEMM('N','T', QLen, RLen, 1,
     $                  FP_One, C(QStart, I), NAO, C(RStart, J), NAO,
     $                  FP_Zero, Dbl_MB( D(QStart, RStart)), NAO)
C
                     Call DGEMM('N','T', QLen, SLen, 1,
     $                  FP_One, C(QStart, I), NAO, C(SStart, J), NAO,
     $                  FP_Zero, Dbl_MB( D(QStart, SStart)), NAO)
C
                     Call DGEMM('N','T', RLen, PLen, 1,
     $                  FP_One, C(RStart, I), NAO, C(PStart, J), NAO,
     $                  FP_Zero, Dbl_MB( D(RStart, PStart)), NAO)
C
                     Call DGEMM('N','T', SLen, PLen, 1,
     $                  FP_One, C(SStart, I), NAO, C(PStart, J), NAO,
     $                  FP_Zero, Dbl_MB( D(SStart, PStart)), NAO)
C
                     Call DGEMM('N','T', RLen, QLen, 1,
     $                  FP_One, C(RStart, I), NAO, C(QStart, J), NAO,
     $                  FP_Zero, Dbl_MB( D(RStart, QStart)), NAO)
C
                     Call DGEMM('N','T', SLen, QLen, 1,
     $                  FP_One, C(SStart, I), NAO, C(QStart, J), NAO,
     $                  FP_Zero, Dbl_MB( D(SStart, QStart)), NAO)
                     T_Den = T_Den + TCGTime()
C
C                    Use this "density" to produce the IJ exchange 
C                    operator.  This requires 8 FP multiply+adds
C                    per integral.
C
                     FPMO = FPMO + 8 * NInts
                     FPAO = FPAO + 8 * NInts
C                    
C                    Prepare to accumulate contributions to this
C                    K_{ij}.
C
                     T_KOp = T_KOp - TCGTime()
                     Call DVZER( NAO*NAO, Dbl_MB( LclK(1, 1)), 1)
C
                     Do N = 1, NInts
                        P = Int_MB( Idx1( N ) )
                        Q = Int_MB( Idx2( N ) )
                        R = Int_MB( Idx3( N ) )
                        S = Int_MB( Idx4( N ) )
C
C                       Consider all permutations of the indices
C                       for possible contributions to KOp
C
                        Dbl_MB( LclK(P, R) ) = Dbl_MB( LclK(P, R))
     $                     + Dbl_MB( D(Q, S)) * Dbl_MB( ERI( N ) )
C                          
                        Dbl_MB( LclK(Q, R)) = Dbl_MB( LclK(Q, R))
     $                     + Dbl_MB( D(P, S)) * Dbl_MB( ERI( N ) )
C
                        Dbl_MB( LclK(P, S)) = Dbl_MB( LclK(P, S))
     $                     + Dbl_MB( D(Q, R)) * Dbl_MB( ERI( N ) )
C
                        Dbl_MB( LclK(Q, S)) = Dbl_MB( LclK(Q, S))
     $                     + Dbl_MB( D(P, R)) * Dbl_MB( ERI( N ) )
C
C
C
                        Dbl_MB( LclK(R, P)) = Dbl_MB( LclK(R, P))
     $                     + Dbl_MB( D(S, Q)) * Dbl_MB( ERI( N ) )
C
                        Dbl_MB( LclK(S, P)) = Dbl_MB( LclK(S, P))
     $                     + Dbl_MB( D(R, Q)) * Dbl_MB( ERI( N ) )
C
                        Dbl_MB( LclK(R, Q)) = Dbl_MB( LclK(R, Q))
     $                     + Dbl_MB( D(S, P)) * Dbl_MB( ERI( N ) )
C
                        Dbl_MB( LclK(S, Q)) = Dbl_MB( LclK(S, Q))
     $                     + Dbl_MB( D(R, P)) * Dbl_MB( ERI( N ) )
C
                     EndDo ! N
                     T_KOp = T_KOp + TCGTime()
C
C                    Accumulate local contribution into global K_{ij}
C                       
                     Count = Count + 1
                     T_Upd = T_Upd - TCGTime()
                     Call GA_Acc( KOp_Hdl, 1, NAO*NAO, IJ, IJ,
     $                  Dbl_MB( LclK(1, 1)), NAO*NAO, FP_One)
                     T_Upd = T_Upd + TCGTime()
                  EndDo ! IJ
C
               EndDo ! SShell
            EndDo    ! RShell
         EndDo       ! QShell
      EndDo          ! PShell
C
C     Free up dynamically allocated work space
C
      LResult = MA_Pop_Stack( LclK_Hdl)
      LResult = MA_Pop_Stack( D_Hdl)
      LResult = Pop_TwoInt()
c$$$C
c$$$      Write (6, *) 'T_Int = ', T_Int
c$$$      Write (6, *) 'T_Den = ', T_Den
c$$$      Write (6, *) 'T_KOp = ', T_KOp
c$$$      Write (6, *) 'T_Upd = ', T_Upd
c$$$      Write (6, *) 'Count = ', Count
C
      Return
      End
