C:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
C NAME
C     Par1_Next_Amp -- Construct next approximation to the amplitudes
C
C SYNOPSIS
      Subroutine Par1_Next_Amp( NExt, NInt, Damp, Step_Hdl, Q, LDQ,
     $   T2_Hdl, MxNorm, FNorm, Scratch, FPMO, FPAO)
      Implicit NONE
      Integer NInt, NExt, LDQ, Step_Hdl, T2_Hdl
      Double Precision Damp, MxNorm, FNorm, FPMO, FPAO
      Double Precision
     $   Q(LDQ, NExt),
     $   Scratch(NExt, NExt)
C
C ARGUMENTS
C     NExt     Dimension of the external space (IN)
C     NInt     Dimension of the internal (occupied) space (IN)
C     Damp     Damping factor, T2(out) = T2(in) + Damp * Step (IN)
C     Step_Hdl GA handle for computed update (residual with denominators) (IN)
C     Q        Projection operator to eliminate contributions
C              due to internal orbtials (IN)
C     LDQ      Leading dimension of Q (IN)
C     T2_Hdl   GA handle for amplitudes (INOUT)
C     MxNorm   Largest elementwise change in amplitudes (OUT)
C     FNorm    F-norm of the change in amplitudes (square root of the
C              sum of the squares of the elements) (OUT)
C     Scratch  A scratch matrix (Scratch)
C     FPMO     Number of FP multiply operations (OUT)
C     FPAO     Number of FP add operations (OUT)
C
C DESCRIPTION
C     Uses the computed update and amplitudes to compute a new set of
C     amplitudes for the next iteration of the equations.  The resulting
C     amplitudes are projected to eliminate contributions from the
C     internal space.  Also returns information about the size of the
C     step actually taken.
C
C     At present, only simple static damping is available.  Eventually
C     this routine will probably involve some kind of conjugate
C     gradient or other approach.
C     
C     Every IJ pair is independent in this routine.
C
C NOTES
C     Perhaps it would be useful to provide convergence information
C     for each IJ pair?
C
C BUGS
C     DNRM2 does fancy scaling to avoid overflow, but the operation
C     count considers only the simplest possible implementation.
C:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#include "numerical_constants.fh"
#include "precision.h"
#include "mafdecls.h"
#include "stdio.fh"
#ifdef TCGMSG
#include "tcgmsg.h"
#include "msgids.fh"
#endif /* TCGMSG */
C FUNCTIONS
      Double Precision DNRM2, DGEAMX
      External DNRM2, DGEAMX
      Logical Verify_Locality
      External Verify Locality
C
C LOCAL VARIABLES
      Integer IJ
      Integer LclT_Hdl, LclT_Ind, LclS_Hdl, LclS_Ind
      Logical LResult
      Integer ExtLo, ExtHi, IJLo, IJHi, Junk
C
      FPMO = FP_Zero
      FPAO = FP_Zero
C
C     **************************
C     * Scale step for damping *
C     **************************
C
      FPMO = FPMO + NExt * NExt * NInt * (NInt+1) / 2
C
      FNorm = FP_Zero
      MxNorm = FP_Zero
C
C     Local buffer for T2, step
C
      LResult = MA_Push_Get( MT_Dbl, NExt*NExt, 'Local T2 buffer',
     $   LclT_Hdl, LclT_Ind)
C
      LResult = MA_Push_Get( MT_Dbl, NExt*NExt, 'Local step buffer',
     $   LclS_Hdl, LclS_Ind)
C
C     ********************************************
C     * Work on the step elements that are local * 
C     ********************************************
C
      Call GA_Distribution( Step_Hdl, NodeID(), ExtLo, ExtHi,
     $   IJLo, IJHi, Junk)
C
C     Check that T2 conforms
C
      If ( .NOT. Verify_Locality( T2_Hdl, ExtLo, ExtHi, IJLo, IJHi) )
     $   Write (LuErr, 9000) 'T2', 'Step',  NodeID()
C
 9000 Format( 1X, 'Par1_Next_Amp: WARNING: Distribution of global ',
     $   'array ', A, ' does not match ', A, 'on process ', I5, '.')
C
      Do IJ = IJLo, IJHi
C
C        *******************************************
C        * Get local copies of amplitudes and step *
C        *******************************************
C
         Call GA_Get( T2_Hdl, 1, NExt*NExt, IJ, IJ,
     $      Dbl_MB(LclT_Ind), NExt*NExt)
         Call GA_Get( Step_Hdl, 1, NExt*NExt, IJ, IJ,
     $      Dbl_MB(LclS_Ind), NExt*NExt)
C
C        *************************************
C        * Scale step to improve convergence *
C        *************************************
C
         Call DSCAL( NExt*NExt, Damp, Dbl_MB( LclS_Ind), 1)
C
C        ***********************************
C        * Collect convergence information *
C        ***********************************
C        Only comparisons in this function, no FP operations.
C
         MxNorm = Max( MxNorm,
     $      DGEAMX( NExt, NExt, Dbl_MB( LclS_Ind), NExt) )
C
C        DNRM2 uses scaling, etc., so the operation count
c        here is likely wrong.
C
         FPMO = FPMO + NExt * NExt
         FPAO = FPAO + NExt * NExt
C
         FNorm = FNorm + DNRM2( NExt*NExt, Dbl_MB( LclS_Ind), 1)**2
C
C        *********************
C        * Update amplitudes *
C        *********************
C
         FPAO = FPAO + NExt * NExt
C
         Call DGEMA('N', NExt, NExt,
     $      FP_One, Dbl_MB( LclS_Ind), NExt,
     $      FP_One, Dbl_MB(LclT_Ind), NExt)
C
C        ***********************************
C        * Project amplitudes to insure    *
C        * orthogonality to internal space *
C        ***********************************
C
         FPMO = FPMO + 2 * NExt * NExt * NExt
         FPAO = FPAO + 2 * NExt * NExt * NExt
C
         Call DGEMM( 'T', 'N', NExt, NExt, NExt,
     $      FP_One, Q, LDQ, Dbl_MB(LclT_Ind), NExt,
     $      FP_Zero, Scratch, NExt)
         Call DGEMM( 'N', 'N', NExt, NExt, NExt,
     $      FP_One, Scratch, NExt, Q, LDQ,
     $      FP_Zero, Dbl_MB(LclT_Ind), NExt)
C
C        **************************************
C        * Put final result back into full T2 *
C        **************************************
C
         Call GA_Put( T2_Hdl, 1, NExt*NExt, IJ, IJ,
     $      Dbl_MB(LclT_Ind), NExt*NExt)
C            
      EndDo ! IJ
C
      LResult = MA_Pop_Stack( LclS_Hdl)
      LResult = MA_Pop_Stack( LclT_Hdl)
C
C     ******************************************
C     * Merge everyone's results for the norms *
C     ******************************************
C
      Call DGOp(MSG_FNorm, FNorm, 1, '+')
      FNorm = Sqrt( FNorm )
C
      Call DGOp(MSG_MxNorm, MxNorm, 1, 'max')
C
      Return
      End
