C:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
C NAME
C     Par1_SolveT2 -- Solve for amplitudes
C
C SYNOPSIS
      Logical Function Par1_SolveT2( NExt, NInt, KOp_Hdl, FInt, LDFInt,
     $   FExt, LDFExt, FQ, LDFQ, Q, LDQ, S, LDS, SInv, LDSInv,
     $   Conv_Energy, Conv_MaxT, Conv_NrmT, MxIter, Damp, T2_Hdl,
     $   Resid_Hdl, Scratch, FPMO, FPAO)
      Implicit NONE
      Integer NInt, NExt, LDFInt, LDFExt, LDFQ, LDQ, LDS, LDSInv, MxIter
      Integer KOp_Hdl, T2_Hdl, Resid_Hdl
      Double precision Conv_Energy, Conv_MaxT, Conv_NrmT, Damp,
     $   FPMO, FPAO
      Double Precision 
     $   FInt(LDFInt, NExt), FExt(LDFExt, NExt), FQ(LDFQ, NExt),
     $   Q(LDQ, NExt), S(LDS, NExt), SInv(LDSInv, NExt),
     $   Scratch(NExt, NExt)
C
C ARGUMENTS
C     NExt    Dimension of the external orbital space (IN)
C     NInt    Dimension of the internal (occupied) orbital space (IN)
C     KOp_Hdl GA handle for exchange operators (IN)
C     FInt    Fock matrix in internal (occupied) basis (IN)
C     LDFInt  Leading dimension of FInt (IN)
C     FExt    Fock matrix in external (usu. AO) basis (IN)
C     LDFExt  Leading dimension of FExt (IN)
C     FQ      Fock matrix in external basis projected to eliminate
C             contributions from internal orbitals (IN)
C     LDFQ    Leading dimension of FQ (IN)
C     Q       Projection operator to remove contributions due
C             to internal orbitals (IN)
C     LDQ     Leading dimension of Q (IN)
C     S       Overlap matrix in external basis (IN)
C     LDS     Leading dimension of S (IN)
C     SInv    Inverse of S (IN)
C     LDSInv  Leading dimension of SInv (IN)
C     Conv_Energy  Largest change in energy allowed at convergence (IN)
C     Conv_MaxT    Largest elementwise change in amplitudes allowed
C                  at convergence (IN)
C     Conv_NrmT    Largest change in amplitudes allowed at convergence
C                  as measured by the F-norm of the step (IN)
C     MxIter       Maximum number of iterations allowed (IN)
C     Damp    Damping factor for Par1_Next_Amp (IN)
C     T2_Hdl  GA handle for T2 amplitudes (OUT)
C     Resid_Hdl    GA handle for scratch array used for residuals 
C                  during iteration (Scratch)
C     Scratch General use scratch array (Scratch)
C     FPMO    Number of FP multiply operations (OUT)
C     FPAO    Number of FP add operations (OUT)
C
C DESCRIPTION
C     Driver for solution of the T2 amplitude equations.  Returns
C     .TRUE. if converged; .FALSE. if MxIter is exceeded.  Convergence
C     is currently defined as satisfying all three of CONV_ENERGY,
C     CONV_MAXT, and CONV_NRMT.  A one-line summary of the
C     correlation energy and the three convergence measures is
C     written to stdout each iteration.
C
C NOTES
C     The F-norm of a matrix grows as the matrix increases in size 
C     (unless all added elements are zero).  Consequently using it
C     for convergence testing gives a criteria which becomes tougher
C     as the size of the problem increases.  So we should probably
C     at least scale it according to the number of amplitudes in the
C     problem.  Perhaps there are even better ways to measure convergence
C     of very large systems of equations?
C:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#include "numerical_constants.fh"
#include "stdio.fh"
#include "precision.h"
#include "mafdecls.h"
#ifdef TCGMSG
#include "tcgmsg.h"
#endif /* TCGMSG */
C FUNCTIONS
      Double Precision Par1_GenE2
      External Par1_GenE2
C
C LOCAL VARIABLES
      Double Precision NrmDiff, E2, FPMO1, FPAO1, MxDiff, LastE2
      Integer Iter
C
C     ****************************
C     * Initialize lots of stuff *
C     ****************************
      Iter = 0
      LastE2 = FP_Zero
C
      FPMO = FP_Zero
      FPAO = FP_Zero
C
      Call GA_Zero( T2_Hdl)
      Call GA_Zero( Resid_Hdl)
C
C     ***************************************
C     * Heading for iteration summary table *
C     ***************************************
C
      If (NodeID() .eq. 0)
     $   Write (LuOut, 9000)
 9000 Format(/1X, 'Iter', 2X, '  Correlation Energy', 2X,
     $   '  delta E', 2X, 'Max del t', 2X, '    RMS t')
      Call Flush(LuOut)
C
C     *************************************************************
C     * Top of loop for iterative solution of amplitude equations *
C     *************************************************************
C
 8000 Continue
C
C     *********************
C     * Evaluate residual *
C     *********************
C
      Call Par1_Eval_Resid( NExt, NInt, KOp_Hdl, T2_Hdl, FInt, LDFInt,
     $   FExt, LDFExt, S, LDS, SInv, LDSInv, Scratch, Resid_Hdl,
     $   FPMO1, FPAO1)
      FPMO = FPMO + FPMO1
      FPAO = FPAO + FPAO1
C
C     *****************************
C     * Form update to amplitudes *
C     *****************************
C
      Call Par1_Apply_Denom( NExt, NInt, Resid_Hdl, FInt, LDFInt,
     $   FQ, LDFQ, Q, LDQ, Scratch, FPMO1, FPAO1)
      FPMO = FPMO + FPMO1
      FPAO = FPAO + FPAO1
C
      Call Par1_Next_Amp( NExt, NInt, Damp, Resid_Hdl, Q, LDQ, T2_Hdl,
     $   MxDiff, NrmDiff, Scratch, FPMO1, FPAO1)
      FPMO = FPMO + FPMO1
      FPAO = FPAO + FPAO1
C
C     **********************************************
C     * Compute energy based on current amplitudes *
C     **********************************************
C
      E2 = Par1_GenE2( KOp_Hdl, T2_Hdl, NExt, NInt, FPMO1, FPAO1)
      FPMO = FPMO + FPMO1
      FPAO = FPAO + FPAO1
C
C     *******************************************
C     * Summarize the results of this iteration *
C     *******************************************
      If (NodeId() .eq. 0)
     $   Write (LuOut, 9010) Iter, E2, E2-LastE2, MxDiff, NrmDiff
 9010 Format( 1X, I4, 2X, F20.15, 2X, SP, 1PE9.2, 2X,
     $   S, 1PE8.2, 2X, 1PE8.2)
      Call Flush( LuOut)
C
C     ********************
C     * Test convergence *
C     ********************
C     Successfully finished our work if converged
C
      If ( Abs( E2 - LastE2 ) .lt. Conv_Energy .AND.
     $   MxDiff .lt. Conv_MaxT .AND. NrmDiff .lt. Conv_NrmT) Then
         Par1_SolveT2 = .TRUE.
         Return
      EndIf
C
C     Unsuccessfully finished if exceeded the maximum number of iterations
C
      If ( Iter .eq. MxIter ) then
         Par1_SolveT2 = .FALSE.
         Return
      EndIf
C
C     Otherwise, go around again
C
      LastE2 = E2
      Iter = Iter + 1
      Goto 8000
C
      End



