C:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
C NAME
C     Par2_make_kop -- Form the exchange operators, K_{ij}
C
C SYNOPSIS
      Subroutine Par2_Make_KOp( C, KOp_Hdl, NAO, NOcc, TotInt,
     $   FPMO, FPAO)
      Implicit NONE
      Integer NAO, NOcc, KOp_Hdl
      Double precision TotInt, FPMO, FPAO
      Double Precision C(NAO, NOcc)
C
C ARGUMENTS
C     C       MO coefficients of the occupied orbitals (IN)
C     KOp_Hdl GA handle for exchange operators (OUT)
C     NAO     Number of AO basis functions (IN)
C     NOcc    Number of occupied orbitals (IN)
C     TotInt  Number of AO integrals actually processed (OUT)
C     FPMO    Number of floating point multiply operations done (OUT)
C     FPAO    Number of floating point add operations done (OUT)
C
C DESCRIPTION
C     Form the exchange operators from AO integrals using an O(o^2N^4)
C     algorithm and a single pass through the integrals.
C
C     The resulting operators are NOT symmetric in their AO indices.
C
C MEMORY USAGE
C     Allocates space to generate two-electron integrals on stack (space
C     required given by Mem_TwoInt) plus NAO*NAO MT_DBLs for the "density"
C     matrix used in the transformation.
C
C NOTES
C     At present, no checking is done of sparsity in the integrals or in
C     the "density", although some sparsity would be expected in both
C     for large systems.  Integrals could be screened using Schwartz's
C     inequality.  For the densities it is a little tougher, since
C     there are actually NOcc*(NOcc+1)/2 different densities, more space
C     would have to be devoted to the screening matrices than might be
C     desirable.  Alternatively, one could implement the screening based
C     on the largest value in each shell block across all of the
C     NOcc*(NOcc+1)/2 densities.  This will probably represent a significant
C     reduction in the sparsity, but may still help a little.
C
C REVISION
C     $Id: par2_make_kop.F,v 1.1 1993/11/08 19:50:00 gg502 Exp $
C:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#include "stdio.fh"
#include "integral_api.fh"
#include "numerical_constants.fh"
#include "precision.h"
#ifdef TCGMSG
#include "tcgmsg.h"
#endif /* TCGMSG */
C FUNCTIONS
      Double Precision DDOT, DNRM2
      External DDOT, DNRM2
C
C LOCAL VARIABLES
C
C     For integral evaluation & other dynamically allocated
C     work space.
C
      Integer IEnv, Val_Ind, P_Ind, Q_Ind, R_Ind, S_Ind, MxNInt,
     $    Scr_Ind, LScratch, D_Hdl, D_Ind, LclK_Hdl, LclK_Ind
      Logical Canonical_Ints, Nonzero_Ints, LResult
C
C     Indices, etc. required for processing integrals
C
      Integer I, J, P, Q, IJ, N, NInts
      Logical Perm_Sym
      Integer NAtom, NShell, NBF, NPrim, MaxPrim, MaxL, Max_Sh_Bf,
     $   Max_At_Bf
C
C     This is unused right now other than to match argument lists 
C     RJH's GTO_* routines.
C
      Integer Basis
C
C     Shell indices and info about basis functions in shell
C
      Integer P_At, Q_At, R_At, S_At, S_AtHi
      Integer P_Sh_Lo, Q_Sh_Lo, R_Sh_Lo, S_Sh_Lo
      Integer P_Sh_Hi, Q_Sh_Hi, R_Sh_Hi, S_Sh_Hi
C
C
C     For determining the local K_{ij}
C
      Integer IJLo, IJHi, Junk
C
C     For load (im)balancing
C
      Integer PQRS, Me, NProc
C
C     Timings & Operation counts
C
      Double precision T_Int, T_Den, T_KOp, T_Upd
      Double precision FPMO1, FPAO1
      Integer Count
C
C STATEMENT FUNCTIONS
C     These statement functions are intended to hide a little of the
C     addressing complexity involved in locally allocated arrays.
C     These functions return appropriate offsets for use within DBL_MB
C     or INT_MB.
C
      Integer ERI, D, LclK
      Integer Idx1, Idx2, Idx3, Idx4
C
      ERI(  N ) = Val_Ind + N - 1
C
      Idx1( N ) = P_Ind + N - 1
      Idx2( N ) = Q_Ind + N - 1
      Idx3( N ) = R_Ind + N - 1
      Idx4( N ) = S_Ind + N - 1
C
      D(P, Q) = D_Ind + P + NAO*(Q-1) - 1
c$$$      LclK(P, Q) = LclK_Ind + P + NAO*(Q-1) - 1
      LclK(P, Q, IJ) = LclK_Ind + P + NAO*(Q-1) - 1 + NAO*NAO*(IJ-1)
C
C     ***********************
C     * Zero out statistics *
C     ***********************
C
      TotInt = FP_Zero
      FPMO = FP_Zero
      FPAO = FP_Zero
C
      T_Int = FP_Zero
      T_Den = FP_Zero
      T_KOp = FP_Zero
      T_Upd = FP_Zero
      Count = 0
C
C     **********************************
C     * Prepare KOp to accumulate data *
C     **********************************
C
      Call GA_Zero( KOp_Hdl)
C
C     ********************************************
C     * Find out what part of KOp we should make *
C     ********************************************
C
c$$$      Call GA_Distribution( KOp_Hdl, NodeID(), AOLo, AOHi, IJLo, IJHi,
c$$$     $   Junk)
C
      Me = NodeID()
      NProc = NNodes()
C
      IJLo = 1
      IJHi = NOcc * (NOcc+1)/2
C
C     If there is no work to do, we can quit early.  Note that
C     all MA allocations occur AFTER this point, so we will skip
C     both the allocation and deallocation.  No matter when we
C     leave this routine, the stack must end up looking as it did
C     on entry.
C
      If (IJLo .eq. 0 .AND. IJHi .eq. -1) then
         Write (LuOut, *) 'Par2_Make_KOp: No work for process ', me
         Return
      EndIf
C
C     *******************************************************
C     * Allocate memory for integral routines and "density" *
C     *******************************************************
C
      LResult = Push_TwoInt_At()
      LResult = Get_Index_TwoInt( IEnv, Val_Ind, P_Ind, Q_Ind, R_Ind,
     $    S_Ind, MxNInt, Scr_Ind, LScratch)
      If ( .NOT. LResult ) Then
         Write (LuErr, *) ' Bad news trying to get indices.'
         STOP
      EndIf
C
      LResult = MA_Push_Get( MT_Dbl, NAO*NAO, 'Density for KOp',
     $   D_Hdl, D_Ind)
      LResult = MA_Push_Get( MT_Dbl, NAO*NAO*NOcc*(NOcc+1)/2,
     $   'Local matrix KOp', LclK_Hdl, LclK_Ind)
C                 
c$$$      Call DVZER( NAO*NAO*NOcc*(NOcc+1)/2, Dbl_MB( LclK(1, 1, 1)), 1)
C
C     ***********************************************************
C     * Get information about the basis for integral evaluation *
C     ***********************************************************
C
      Call GTO_Info( Basis, NAtom, NShell, NBF, NPrim, MaxPrim,
     $   MaxL, Max_Sh_Bf, Max_At_Bf)
C
C     ********************************************
C     * Control massaging of integrals by TwoInt *
C     ********************************************
C     The indices need not be completely canonical.
C     Better to remove zero integrals once than process them many times???
C
      Canonical_Ints = .FALSE.
      Nonzero_Ints = .TRUE.
      Perm_Sym = .TRUE.
C
C     *****************************************
C     * Loop unique shells in canonical order *
C     *****************************************
C
      PQRS = 0
      Do P_At = 1, NAtom
         Call GTO_At_to_Sh( Basis, P_At, P_Sh_Lo, P_Sh_Hi)
C
         Do Q_At = 1, P_At
            Call GTO_At_to_Sh( Basis, Q_At, Q_Sh_Lo, Q_Sh_Hi)
C
            Do R_At = 1, P_At
               Call GTO_At_to_Sh( Basis, R_At, R_Sh_Lo, R_Sh_Hi)
C
               S_AtHi = R_At
               If (P_At .eq. R_At) S_AtHi = Q_At
C
               Do S_At = 1, S_AtHi
                  Call GTO_At_to_Sh( Basis, S_At, S_Sh_Lo, S_Sh_Hi)
C
                  If ( Mod( PQRS, NProc) .eq. Me ) then
C
C                    Generate a batch of integrals
C
                     T_Int = T_Int - TCGTime()
                     Call TwoInt_At_PermSym(IEnv, Basis,
     $                  P_At, Q_At, R_At, S_At,
     $                  DBL_MB( ERI(1)),
     $                  Int_MB( Idx1(1)),
     $                  Int_MB( Idx2(1)),
     $                  Int_MB( Idx3(1)),
     $                  Int_MB( Idx4(1)),
     $                  NInts, Dbl_MB(Scr_Ind),
     $                  LScratch, Canonical_Ints, Nonzero_Ints,
     $                  Perm_Sym, FPMO1, FPAO1)
                     T_Int = T_Int + TCGTime()
C
                     TotInt = TotInt + NInts
                     FPMO = FPMO + FPMO1
                     FPAO = FPAO + FPAO1
C
                     Do IJ = IJLo, IJHi
                        Count = Count + 1
                        Call Detriangularize(IJ, I, J)
C
C                       Form a density-like beast, D_{\mu \nu}^{ij}
C                       from the I and J columns of the MO coefficients.
C                       Note that this matrix is NOT symmetric!
C
                        T_Den = T_Den - TCGTime()
C
                        Call Mk_Dij_AtQt( NAO, C, I, J, P_At, Q_At,
     $                     R_At, S_At, Dbl_MB( D(1, 1)), FPMO1,
     $                     FPAO1)
C
                        FPMO = FPMO + FPMO1
                        FPAO = FPAO + FPAO1
C
                        T_Den = T_Den + TCGTime()
C
C                       Contract the integrals with the densities
C                       to form some exchange operators
C
                        T_KOp = T_KOp - TCGTime()
C
                        Call DVZER( NAO*NAO, Dbl_MB( LclK(1, 1, IJ)), 1)
C
                        Call Par2_Int_into_KOp(NInts,
     $                     Int_MB(Idx1(1)),
     $                     Int_MB(Idx2(1)),
     $                     Int_MB(Idx3(1)),
     $                     Int_MB(Idx4(1)),
     $                     Dbl_MB(ERI(1)),
     $                     NAO, 1, Dbl_MB( D(1, 1)),
     $                     Dbl_MB( LclK(1, 1, IJ)), FPMO1, FPAO1)
                        FPMO = FPMO + FPMO1
                        FPAO = FPAO + FPAO1
C
                        T_KOp = T_KOp + TCGTime()
C
C                       Accumulate local contribution into global K_{ij}
C                       
                        T_Upd = T_Upd - TCGTime()
                        Call GA_Acc( KOp_Hdl, 1, NAO*NAO, IJ, IJ,
     $                     Dbl_MB( LclK(1, 1, IJ)), NAO*NAO, FP_One)
                        T_Upd = T_Upd + TCGTime()
C
                     EndDo      ! IJ
C
                  EndIf
C
C                 Update counter for load (im)balancing
C
                  PQRS = PQRS + 1
               EndDo ! S_At
            EndDo    ! R_At
         EndDo       ! Q_At
      EndDo          ! P_At
C
      Do IJ = IJLo, IJHi
         Call GA_Get(KOp_Hdl, 1, NAO*NAO, IJ, IJ,
     $      Dbl_MB( LclK(1, 1, IJ)), NAO*NAO)
         Write (6, *) 'KOp checksum ', IJ, DNRM2( NAO*NAO,
     $      Dbl_MB( LclK(1, 1, IJ)), 1)
      EndDo
c$$$C
c$$$C     Accumulate local contribution into global K_{ij}
c$$$C                       
c$$$      T_Upd = T_Upd - TCGTime()
c$$$      Call GA_Acc( KOp_Hdl, 1, NAO*NAO, IJLo, IJHi,
c$$$     $   Dbl_MB( LclK(1, 1, 1)), NAO*NAO, FP_One)
c$$$      T_Upd = T_Upd + TCGTime()
C
C     Free up dynamically allocated work space
C
      LResult = MA_Pop_Stack( LclK_Hdl)
      LResult = MA_Pop_Stack( D_Hdl)
      LResult = Pop_TwoInt_At()
C
      Write (6, *) 'T_Int = ', T_Int
      Write (6, *) 'T_Den = ', T_Den
      Write (6, *) 'T_KOp = ', T_KOp
      Write (6, *) 'T_Upd = ', T_Upd
      Write (6, *) 'Count = ', PQRS, Count
C
      Return
      End
