C:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
C NAME
C     Par3 -- simple-minded E(2) calculations
C
C SYNOPSIS
      Subroutine Par3( NExt, NInt, C_Hdl, FInt_Hdl, FExt_Hdl, S_Hdl,
     $   Q_Hdl, SInv_Hdl, FQ_Hdl)
      Implicit NONE
      Integer NExt, NInt, C_Hdl, FInt_Hdl, FExt_Hdl,
     $   S_Hdl, Q_Hdl, SInv_Hdl, FQ_Hdl
C
C ARGUMENTS
C     NExt     Dimension of external basis (IN)
C     NInt     Dimension of internal (occupied) basis (IN)
C     C_Hdl    Memhandle for SCF MO coefficients (IN)
C     FInt_Hdl Memhandle for Fock matrix in internal orbital basis (IN)
C     FExt_Hdl Memhandle for Fock matrix in external orbtial basis (IN)
C     S_Hdl    Memhandle for external basis overlap matrix (IN)
C     Q_Hdl    Memhandle for projection operator to remove internal
C              orbital contributions from external basis objects (IN)
C     SInv_Hdl Memhandle for S^{-1} (IN)
C     FQ_Hdl   Memhandle for Fock matrix projected with Q (IN)
C
C DESCRIPTION
C     Driver for E(2) energy evaluation.  The method is general -- able to
C     handle both canonical and noncanonical Hartree-Fock references.
C     It can, in principle, operate with all of the relevant quantities
C     in the MO basis, or in a mixture of MO and AO basis (the all-MO 
C     operation has not been verified recently).  More efficient algorithms
C     can be had for special cases -- canonical orbitals, MO basis
C     quantities, etc.
C
C NOTES
C     In this formalism, neither the exchange operators, nor the T amplitudes
C     are in general symmetric (or antisymmetric).  There is a "double
C     transpose" symmetry, so that t_{ij} = t_{ji}^T and likewise for K.
C
C DYNAMIC MEMORY
C     Uses 3 * NExt*Next*NInt*(NInt+1)/2 MT_DBLs for exchange operators,
C     T2 amplitudes, and residuals.  Uses NExt*NExt MT_DBLs for general
C     scratch in solution of T2 amplitude equations.
C
C     Children may also dynamically allocate memory.
C
C BUGS
C     Leading dimensions of matrices are not passed with the arguments.
C
C     Convergence criteria are local parameters -- caller should have
C     control.
C
C REVISION
C     $Id: par3.F,v 1.1 1993/11/08 19:50:07 gg502 Exp $
C:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#include "stdio.fh"
#include "mafdecls.h"
#include "numerical_constants.fh"
#include "precision.h"
#include "global.h"
#ifdef TCGMSG
#include "tcgmsg.h"
#endif /* TCGMSG */
C
C PARAMETERS
      Double precision Conv_Energy, Conv_MaxT, Conv_NrmT
      Parameter (Conv_Energy = 1.0d-6, Conv_MaxT = 1.0d-6,
     $   Conv_NrmT = 1.0d-6)
C
      Integer MxIter
      Parameter (MxIter = 1000)
C
      Double precision Damping
      Parameter (Damping = FP_Half)
C
C FUNCTIONS
      Double Precision Toy1_GenE2, Par3_GenE2
      External Toy1_GenE2, Par3_GenE2
      Logical Par3_SolveT2
      External Par3_SolveT2
C
C LOCAL VARIABLES
      Logical LResult
C
C     Indices corresponding to go with memhandles in argument list
C
      Integer C_Ind, FInt_Ind, FExt_Ind, S_Ind, Q_Ind,
     $   SInv_Ind, FQ_Ind
C
C     Handles and indices for locally dynamically allocated objects
C
      Integer Scratch_Hdl, T2_Hdl, Resid_Hdl, KOp_Hdl
      Integer Scratch_Ind
C
C     Final energy, operation counts, etc.
C
      Double Precision E2, FPMO, FPAO, TotInt
C
#ifdef TCGMSG
C
C     Timing information
C
      Double precision TBegin, TEnd
#endif /* TCGMSG */
C
      Write (LuOut, *) 'Starting Par3 calculations'
C
C     ***************************************
C     * Form the exchange operators, K_{ij} *
C     ***************************************
C     This is basically a two-index transformation:
C     C' (mu, lambda | nu, sigma) C --> (I, lambda | J, sigma)
C
c$$$      LResult = GA_Create( MT_Dbl, NExt*NExt, NInt*(NInt+1)/2,
c$$$     $   'Exchange operators', NExt*NExt, 1, KOp_Hdl)
c$$$C
C     Make sure we have a current index for the coefficients
C
      LResult = MA_Get_Index( C_Hdl, C_Ind)
C
#ifdef TCGMSG
      Call FFFlush( LuOut)
      Call GA_Sync
      TBegin = TCGTime()
#endif /* TCGMSG */
C
      Call Par3_Make_KOp( Dbl_MB( C_Ind), KOp_Hdl,
     $   NExt, NInt, TotInt, FPMO, FPAO)
C
#ifdef TCGMSG
      Call GA_Sync
      TEnd = TCGTime()
#endif /* TCGMSG */
C
      Write (LuOut, 9000) NInt*(NInt+1)/2, TotInt, FPMO, FPAO,
     $  GA_NodeID()
 9000 Format( /1X, 'Formed ', I8, ' exchange operators',
     $   ' from ', 1PE10.3, ' integrals',
     $   /1X, 'using ', 1PE10.3, ' FP* and ', 1PE10.3,
     $   ' FP+ operations on ', I5, '.')
C
#ifdef TCGMSG
      If (GA_NodeID() .eq. 0) Write (LuOut, 9005) 'KOp', TEnd-TBegin
 9005 Format(1X, 'Time for ', A, ': ', F12.3, ' seconds.')
#endif /* TCGMSG */
C
C     *********************************************************
C     * Form T amplitudes, compute energy (mixed MO/AO basis) *
C     *********************************************************
C
      LResult = GA_Create( MT_Dbl, NExt*NExt, NInt*(NInt+1)/2,
     $   'T2 amplitudes', NExt*NExt, 1, T2_Hdl)
C
C     These two are needed only during determination of the amplitudes

      LResult = GA_Create(MT_Dbl, NExt*NExt, NInt*(NInt+1)/2,
     $   'Residual', NExt*NExt, 1, Resid_Hdl)
      LResult = MA_Push_Get(MT_Dbl, NExt*NExt, 'Scratch',
     $   Scratch_Hdl, Scratch_Ind)
C
C     Make sure we have current indices for everything
C
      LResult = MA_Get_Index( FInt_Hdl, FInt_Ind)
      LResult = MA_Get_Index( FExt_Hdl, FExt_Ind)
      LResult = MA_Get_Index( FQ_Hdl, FQ_Ind)
      LResult = MA_Get_Index( Q_Hdl, Q_Ind)
      LResult = MA_Get_Index( S_Hdl, S_Ind)
      LResult = MA_Get_Index( SInv_Hdl, SInv_Ind)
C
#ifdef TCGMSG
      Call FFFlush( LuOut)
      Call GA_Sync
      TBegin = TCGTime()
#endif /* TCGMSG */
C
      LResult = Par3_SolveT2( NExt, NInt, KOp_Hdl,
     $   Dbl_MB(FInt_Ind), NExt, Dbl_MB(FExt_Ind), NExt,
     $   Dbl_MB(FQ_Ind), NExt, Dbl_MB(Q_Ind), NExt,
     $   Dbl_MB(S_Ind), NExt, Dbl_MB(SInv_Ind), NExt,
     $   Conv_Energy, Conv_MaxT, Conv_NrmT, MxIter, Damping,
     $   T2_Hdl, Resid_Hdl,
     $   Dbl_MB(Scratch_Ind), FPMO, FPAO )
C
#ifdef TCGMSG
      Call GA_Sync
      TEnd = TCGTime()
#endif /* TCGMSG */
C
      If (LResult ) then
         Write (LuOut, 9020) 'Converged amplitudes',  FPMO, FPAO,
     $      GA_NodeID()
      Else
         Write (LuOut, *) ' Par3: Amplitudes failed to converge.'
         Stop
      EndIf
 9020 Format( 1X, A, ' required ', 1PE10.3, ' FP* and ',
     $   1PE10.3, ' FP+ operations on ', I5, '.')
C
#ifdef TCGMSG
      If (GA_NodeID() .eq. 0) Write (LuOut, 9005) 'amplitudes',
     $  TEnd-TBegin
#endif /* TCGMSG */
C
C     Get rid of the temporaries
C
      LResult = MA_Pop_Stack( Scratch_Hdl)
      LResult = GA_Destroy( Resid_Hdl)
C
#ifdef TCGMSG
      Call FFFlush( LuOut)
      Call GA_Sync
      TBegin = TCGTime()
#endif /* TCGMSG */
C
      E2 = Par3_GenE2( KOp_Hdl, T2_Hdl,
     $   NExt, NInt, FPMO, FPAO)
C
#ifdef TCGMSG
      Call GA_Sync
      TEnd = TCGTime()
#endif /* TCGMSG */
C
      Write (LuOut, *) 'E(2) = ', E2
C
      Write (LuOut, 9010) FPMO, FPAO, GA_NodeID()
 9010 Format( 1X, 'Energy evaluation required ', 1PE10.3, ' FP* and ',
     $   1PE10.3, ' FP+ operations on ', I5, '.')
C
#ifdef TCGMSG 
      If (GA_NodeID() .eq. 0) Write (LuOut, 9005) 'energy', TEnd-TBegin
#endif /* TCGMSG */
C     
C     ***************************
C     * Cleanup after ourselves *
C     ***************************
C
      LResult = GA_Destroy( T2_Hdl )
      LResult = GA_Destroy( KOp_Hdl)
C
      Write (LuOut, *) 'Done with Par3 calculations'
C
      Return
      End
