C:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
C NAME
C     Par3_Eval_Resid -- Evaluate the residual of the amplitude equations
C
C SYNOPSIS
      Subroutine Par3_Eval_Resid( NExt, NInt, KOp_Hdl, T2_Hdl,
     $   FInt, LDFInt, FExt, LDFExt, S, LDS, SInv, LDSInv, Scratch,
     $   R_Hdl, FPMO, FPAO)
      Implicit NONE
      Integer NExt, NInt, LDFInt, LDFExt, LDS, LDSInv
      Integer KOp_Hdl, T2_Hdl, R_Hdl
      Double precision FPMO, FPAO
      Double Precision 
     $   FInt(LDFInt, NInt), FExt(LDFExt, NExt), S(LDS, NExt),
     $   SInv(LDSInv, NExt), Scratch(NExt, NExt)
C
C ARGUMENTS
C     NExt    Dimension of the external orbital space (IN)
C     NInt    Dimension of the internal (occupied) orbital space (IN)
C     KOp_Hdl GA handle for exchange operators (IN)
C     T2_Hdl  GA handle for amplitudes to be used for evaluation of 
C             residual (IN)
C     FInt    Fock matrix in internal (occupied) basis (IN)
C     LDFInt  Leading dimension of FInt (IN)
C     FExt    Fock matrix in external (usu. AO) basis (IN)
C     LDFExt  Leading dimension of FExt (IN)
C     S       Overlap matrix in external basis (IN)
C     LDS     Leading dimension of S (IN)
C     SInv    Inverse of S (IN)
C     LDSInv  Leading dimension of SInv (IN)
C     Scratch An NExt*NExt scratch matrix (Scratch)
C     R_Hdl   GA handle for residual in external basis (OUT)
C     FPMO    Number of FP multiply operations (OUT)
C     FPAO    Number of FP add operations (OUT)
C
C DESCRIPTION
C     Evaluate the "residual", r_{ij}:
C     S r_{ij} S = K_{ij} - K_{ij}^\dagger + F t_{ij} S + S t_{ij} F 
C                - S ( f_{im} t_{mj} + t_{im} f_{mj} ) S
C
C     For efficiency, this is implemented as:
C     r_{ij} = S^{-1} [ K_{ij} - K_{ij}^\dagger + F t_{ij} S 
C            + S t_{ij} F ] S^{-1} - f_{im} t_{mj} + t_{im} f_{mj}
C     which is 2 matrix multiplications per $ij$ cheaper.
C
C     This routine is roughly O( 6 NExt^3 NInt(NInt+1)/2 ) FP
C     multiplications and additions.  The actual numbers are accumulated
C     and returned in FPMO and FPAO.
C
C NOTES
C     An optimization for canonical internal orbitals, or more
C     generally for small f * t would help reduce the work in the
C     last two terms (the four loops on M, below).  The general version
C     would require some input about the size of t_{ij}, but this
C     could easily be evaluated as bart of the convergence test.
C     For storage efficiency, it could even be compressed.
C
C     An optimization for the case of orthogonal external orbitals
C     would be to avoid multiplication by S and SInv.  The problem
C     is that this would really require a different routine, since
C     the data movement would not be optimal (copy operations between
C     Scratch and R would have to be done instead of the multiplications.
C
C     In a similar vein, perhaps it would be useful for the caller to
C     pass arguments indicating if either (internal or external) Fock
C     matrices are diagonal in order to bypass some operations.
C     Once again, this starts to be a different routine pretty quickly.
C:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#include "numerical_constants.fh"
#include "precision.h"
#include "mafdecls.h"
#include "stdio.fh"
#include "global.h"
#ifdef TCGMSG
#include "tcgmsg.h"
#endif /* TCGMSG */
C FUNCTIONS
      Logical Verify_Locality
      External Verify_Locality
C     
C LOCAL VARIABLES
      Integer I, J, M, IJ, IM, MJ
      Integer LclK_Hdl, LclK_Ind, LclT_Hdl, LclT_Ind, LclR_Hdl, LclR_Ind
      Logical LResult
      Integer PID, ExtLo, ExtHi, IJLo, IJHi, Junk
C
      PID = GA_NodeID()
C
C     Allocate local buffers
C
      LResult = MA_Push_Get( MT_Dbl, NExt*NExt, 'Local KOp buffer',
     $   LclK_Hdl, LclK_Ind)
      LResult = MA_Push_Get( MT_Dbl, NExt*NExt, 'Local T2 buffer',
     $   LclT_Hdl, LclT_Ind)
      LResult = MA_Push_Get( MT_Dbl, NExt*NExt, 'Local residual buffer',
     $   LclR_Hdl, LclR_Ind)
C
      FPMO = FP_Zero
      FPAO = FP_Zero
C
C     *************************************************
C     * Generate the residual elements that are local * 
C     *************************************************
C
      Call GA_Distribution( R_Hdl, PID, ExtLo, ExtHi, IJLo, IJHi, Junk)
C
C     Check that KOp and T2 conform
C
      If ( .NOT. Verify_Locality( T2_Hdl, ExtLo, ExtHi, IJLo, IJHi) )
     $   Write (LuErr, 9000) 'T2', 'R',  PID
      If ( .NOT. Verify_Locality( KOp_Hdl, ExtLo, ExtHi, IJLo, IJHi) )
     $   Write (LuErr, 9000) 'KOp', 'R', PID
C
 9000 Format( 1X, 'Par3_Eval_Resid: WARNING: Distribution of global ',
     $   'array ', A, ' does not match ', A, 'on process ', I5, '.')
C
      Do IJ = IJLo, IJHi
         Call Detriangularize( IJ, I, J)
C
C        *************************
C        * Get local copy of KOp *
C        *************************
C
         Call GA_Get( KOp_Hdl, 1, NExt*NExt, IJ, IJ,
     $      Dbl_MB( LclK_Ind), NExt*NExt)
C
c        ****************************
C        * Operations for KOp terms *
c        ****************************
C
         FPAO = FPAO + 3 * NExt*NExt
C
C        ********************************************************
C        * Alpha-alpha + beta-beta integral (KOp) contributions *
C        ********************************************************
C           
         Call DGEMA('N', NExt, NExt,
     $      FP_One, Dbl_MB(LclK_Ind), NExt,
     $      FP_Zero, Dbl_MB( LclR_Ind), NExt)
         Call DGEMA('T', NExt, NExt,
     $      -FP_One, Dbl_MB(LclK_Ind), NExt,
     $      FP_One, Dbl_MB( LclR_Ind), NExt)
C           
C        ***************************
C        * Alpha-beta contribution *
C        ***************************
C
         Call DGEMA('N', NExt, NExt,
     $      FP_One, Dbl_MB( LclK_Ind), NExt,
     $      FP_One, Dbl_MB( LclR_Ind), NExt)
C
C        ************************
C        * Get local copy of T2 *
C        ************************
C
         Call GA_Get( T2_Hdl, 1, NExt*NExt, IJ, IJ,
     $      Dbl_MB( LclT_Ind), NExt*NExt)
C
C        **************************************
C        * Operations for F t S + S t F terms *
C        **************************************
C
         FPMO = FPMO + 4 * NExt * NExt * NExt
         FPAO = FPAO + 4 * NExt * NExt * NExt
C
C        ***************************************
C        * r(ij,ab) <-- + F(ae) T(ij,ef) S(fb) *
C        *              + S(af) T(ij,fe) F(eb) *
C        ***************************************
C
         Call DGEMM( 'N', 'N', NExt, NExt, NExt, FP_One,
     $      FExt, LDFExt, Dbl_MB( LclT_Ind), NExt,
     $      FP_Zero, Scratch, NExt)
         Call DGEMM( 'N', 'N', NExt, NExt, NExt, FP_One,
     $      Scratch, NExt, S, LDS, FP_One, Dbl_MB( LclR_Ind), NExt)
C
         Call DGEMM( 'N', 'N', NExt, NExt, NExt, FP_One,
     $      S, LDS, Dbl_MB(LclT_Ind), NExt, FP_Zero, Scratch, NExt)
         Call DGEMM( 'N', 'N', NExt, NExt, NExt, FP_One,
     $      Scratch, NExt, FExt, LDFExt,
     $      FP_One, Dbl_MB( LclR_Ind), NExt)
C
C        **************************************
C        * Operations for back transformation *
C        **************************************
C
         FPMO = FPMO + 2 * NExt * NExt * NExt
         FPAO = FPAO + 2 * NExt * NExt * NExt
C
C        *******************************************************
C        * Transform Scr to AO basis [ S r_{ij} S --> r_{ij} ] *
C        *******************************************************
C        Do this now to avoid having to multiply the inter-pair
C        terms by S then immediately by S^-1.
C
         Call DGEMM( 'N', 'N', NExt, NExt, NExt,
     $      FP_One, SInv, LDSInv, Dbl_MB( LclR_Ind), NExt,
     $      FP_Zero, Scratch, NExt)
         Call DGEMM( 'N', 'N', NExt, NExt, NExt,
     $      FP_One, Scratch, NExt, SInv, LDSInv,
     $      FP_Zero, Dbl_MB( LclR_Ind), NExt)
C
C        *****************************************************
C        * Interaction of other occupied pairs with this one *
C        *****************************************************
C        Note that t_{ji} = t_{ij}^T, so we must break up
C        the range of M in order to do put each piece of t into
C        the summation in the proper orientation.
C
         Do M = 1, J-1
            MJ = Max( M, J) * (Max(M, J) - 1 )/2 + Min(M, J)
C
C           ***********************************
C           * r(ij,ab) <-- -FInt(im) T(mj,ab) *
C           ***********************************
C
            Call GA_Get( T2_Hdl, 1, NExt*NExt, MJ, MJ,
     $         Dbl_MB( LclT_Ind), NExt*NExt)
C
            FPMO = FPMO + NExt * NExt
            FPAO = FPAO + NExt * NExt
C
            Call DGEMA('T', NExt, NExt,
     $         -FInt(I, M), Dbl_MB(LclT_Ind), NExt,
     $         FP_One, Dbl_MB( LclR_Ind), NExt)
         EndDo                  ! M
C
         Do M = J, NInt
            MJ = Max( M, J) * (Max(M, J) - 1 )/2 + Min(M, J)
C
C           ***********************************
C           * r(ij,ab) <-- -FInt(im) T(mj,ab) *
C           ***********************************
C
            Call GA_Get( T2_Hdl, 1, NExt*NExt, MJ, MJ,
     $         Dbl_MB( LclT_Ind), NExt*NExt)
C
            FPMO = FPMO + NExt * NExt
            FPAO = FPAO + NExt * NExt
C
            Call DGEMA('N', NExt, NExt,
     $         -FInt(I, M), Dbl_MB(LclT_Ind), NExt,
     $         FP_One, Dbl_MB( LclR_Ind), NExt)
         EndDo                  ! M
C
         Do M = 1, I
            IM = Max( M, I) * (Max(M, I) - 1 )/2 + Min(M, I)
C
C           ***********************************
C           * r(ij,ab) <-- -T(im,ab) FInt(mj) *
C           ***********************************
C
            Call GA_Get( T2_Hdl, 1, NExt*NExt, IM, IM,
     $         Dbl_MB( LclT_Ind), NExt*NExt)
C
            FPMO = FPMO + NExt * NExt
            FPAO = FPAO + NExt * NExt
C
            Call DGEMA('N', NExt, NExt,
     $         -FInt(M, J), Dbl_MB(LclT_Ind), NExt,
     $         FP_One, Dbl_MB( LclR_Ind), NExt)
         EndDo                  ! M 
C
         Do M = I+1, NInt
            IM = Max( M, I) * (Max(M, I) - 1 )/2 + Min(M, I)
C
C           ***********************************
C           * r(ij,ab) <-- -T(im,ab) FInt(mj) *
C           ***********************************
C
            Call GA_Get( T2_Hdl, 1, NExt*NExt, IM, IM,
     $         Dbl_MB( LclT_Ind), NExt*NExt)
C
            FPMO = FPMO + NExt * NExt
            FPAO = FPAO + NExt * NExt
C
            Call DGEMA('T', NExt, NExt,
     $         -FInt(M, J), Dbl_MB(LclT_Ind), NExt,
     $         FP_One, Dbl_MB( LclR_Ind), NExt)
         EndDo                  ! M 
C
C        ********************************************************
C        * Put this piece of the residual into the global space *
C        ********************************************************
C
         Call GA_Put( R_Hdl, 1, NExt*NExt, IJ, IJ,
     $      Dbl_MB( LclR_Ind), NExt*NExt)
C
      EndDo      ! IJ
C
      LResult = MA_Pop_Stack( LclR_Hdl)
      LResult = MA_Pop_Stack( LclT_Hdl)
      LResult = MA_Pop_Stack( LclK_Hdl)
C
      Return
      End

