C:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
C NAME
C     Par3_make_kop -- Form the exchange operators, K_{ij}
C
C SYNOPSIS
      Subroutine Par3_Make_KOp( C, KOp_Hdl, NAO, NOcc, TotInt,
     $   FPMO, FPAO)
      Implicit NONE
      Integer NAO, NOcc, KOp_Hdl
      Double precision TotInt, FPMO, FPAO
      Double Precision C(NAO, NOcc)
C
C ARGUMENTS
C     C       MO coefficients of the occupied orbitals (IN)
C     KOp_Hdl GA handle for exchange operators (OUT)
C     NAO     Number of AO basis functions (IN)
C     NOcc    Number of occupied orbitals (IN)
C     TotInt  Number of AO integrals actually processed (OUT)
C     FPMO    Number of floating point multiply operations done (OUT)
C     FPAO    Number of floating point add operations done (OUT)
C
C DESCRIPTION
C     Form the exchange operators from AO integrals using an O(o^2N^4)
C     algorithm and a single pass through the integrals.
C
C     The resulting operators are NOT symmetric in their AO indices.
C
C MEMORY USAGE
C     Allocates space to generate two-electron integrals on stack (space
C     required given by Mem_TwoInt) plus NAO*NAO MT_DBLs for the "density"
C     matrix used in the transformation.
C
C NOTES
C     At present, no checking is done of sparsity in the integrals or in
C     the "density", although some sparsity would be expected in both
C     for large systems.  Integrals could be screened using Schwartz's
C     inequality.  For the densities it is a little tougher, since
C     there are actually NOcc*(NOcc+1)/2 different densities, more space
C     would have to be devoted to the screening matrices than might be
C     desirable.  Alternatively, one could implement the screening based
C     on the largest value in each shell block across all of the
C     NOcc*(NOcc+1)/2 densities.  This will probably represent a significant
C     reduction in the sparsity, but may still help a little.
C
C REVISION
C     $Id: par3_make_kop.F,v 1.1 1993/11/08 19:50:25 gg502 Exp $
C:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#include "stdio.fh"
#include "integral_api.fh"
#include "numerical_constants.fh"
#include "precision.h"
#ifdef TCGMSG
#include "tcgmsg.h"
#endif /* TCGMSG */
#include "global.h"
C FUNCTIONS
      Double Precision DDOT, DNRM2, DASUM
      External DDOT, DNRM2, DASUM
      Logical BlockSize
      External BlockSize
      Integer Par3_Grab_Buffer, Par3_Check_Done
      External Par3_Grab_Buffer, Par3_Check_Done
      Integer NxTask
      External NxTask
C
C LOCAL VARIABLES
C
C     For integral evaluation & other dynamically allocated
C     work space.
C
      Integer IEnv, Val_Ind, P_Ind, Q_Ind, R_Ind, S_Ind, MxNInt,
     $    Scr_Ind, LScratch, D_Hdl, D_Ind
      Logical Canonical_Ints, Nonzero_Ints, LResult
C
C     Indices, etc. required for processing integrals
C
      Integer P, Q, IJ, N, NInts
      Logical Perm_Sym
      Double precision Fact
      Integer NAtom, NShell, NBF, NPrim, MaxPrim, MaxL, Max_Sh_Bf,
     $   Max_At_Bf
C
C     This is unused right now other than to match argument lists 
C     RJH's GTO_* routines.
C
      Integer Basis
C
C     Shell indices and info about basis functions in shell
C
      Integer PSh, QSh, RSh, SSh, SShHi
C
C     For determining the local K_{ij}
C
      Integer IJLo, IJHi
C
C     For load (im)balancing
C
      Integer PQRS, Me, NProc
C
C     Timings & Operation counts
C
      Double precision T_Int, T_Den, T_KOp, T_Upd
      Double precision T_Spin, T_Sync
      Double precision FPMO1, FPAO1
      Integer Count
C
      Integer MyKOp_Hdl
      Integer MyKOp_Ind
      Integer RBlkSz, CBlkSz, NProc_Row, Len, MyRowDim, MyColDim
      Integer Me_Row, Me_Col, MyKOp_Row, MyKOp_Col
C
      Integer From
      Integer Ready_Hdl, Processed_Hdl
      Integer Ready_Ind, Processed_Ind
      Integer Free_Buffer, Buf_per_Proc, Total_Buffers, Still_Running,
     $   Last_Running, Not_Free
      Logical Last_Pass
      Integer Int_Batch_Granularity, Next_Batch
      Integer Batches_Prod, Ints_Prod, Probes, No_Buffers, Work
C
C STATEMENT FUNCTIONS
C     These statement functions are intended to hide a little of the
C     addressing complexity involved in locally allocated arrays.
C     These functions return appropriate offsets for use within DBL_MB
C     or INT_MB.
C
      Integer ERI, D
      Integer Idx1, Idx2, Idx3, Idx4
      Integer Which_Proc
C
      ERI(  N ) = Val_Ind + N - 1
C
      Idx1( N ) = P_Ind + N - 1
      Idx2( N ) = Q_Ind + N - 1
      Idx3( N ) = R_Ind + N - 1
      Idx4( N ) = S_Ind + N - 1
C
      D(P, Q) = D_Ind + P + NAO*(Q-1) - 1
C
C     Convert from a basis function pair to a processor number
C     holding the corresponding piece of KOp.
C
C     Note that processors span 0..NProc-1, so I don't need to add
C     1 to the P/RBlkSz term.
C
      Which_Proc( P, Q ) = ( P / RBlkSz ) + NProc_Row * ( Q / CBlkSz )
C
C     ***********************
C     * Zero out statistics *
C     ***********************
C
      TotInt = FP_Zero
      FPMO = FP_Zero
      FPAO = FP_Zero
C
      T_Int = FP_Zero
      T_Den = FP_Zero
      T_KOp = FP_Zero
      T_Upd = FP_Zero
      T_Spin = FP_Zero
      T_Sync = FP_Zero
      Count = 0
      Batches_Prod = 0
      Ints_Prod = 0
C
      Probes = 0
      No_Buffers = 0
      Work = 0
C
C     ***********************************************************
C     * Get information about the basis for integral evaluation *
C     ***********************************************************
C
      Call GTO_Info( Basis, NAtom, NShell, NBF, NPrim, MaxPrim,
     $   MaxL, Max_Sh_Bf, Max_At_Bf)
C
      NProc = GA_NNodes()
      Me = GA_NodeID()
C
C     *************************************************************
C     * Set granularity of integral processing for load balancing *
C     *************************************************************
C     We are using a shared counter for dynamic load balancing.
C     Accessing it takes time, so we want to insure we do enough
C     work on each increment to make the overhead of accessing the
C     shared counter negligible.  Thus we use NxTask which allocates
C     work in blocks of size Int_Batch_Granularity, which we can set here
C     according to the platform or other factors.
C
      Int_Batch_Granularity = 10
C
C     ********************************************
C     * Find out what part of KOp we should make *
C     ********************************************
C     Do all internal pairs on each process.
C
      IJLo = 1
      IJHi = NOcc * (NOcc+1) / 2
C
C     Divide the external space evenly across processors.  Using
C     rectangular blocks should make it more likely that a bunch 
C     of integrals in a given shell block will be useful.
C
C     An eventual optimization would be to recognize that every integral
C     contributing to the (P,R) element of the exchange operators
C     will also contribute to the (R,P) element and keep the element
C     and its transpose together on the same node.  This would reduce
C     communication requirements by 2x.
C
      LResult = BlockSize( NProc, NAO, NAO, 1, 1, RBlkSz, CBlkSz)
      If ( .NOT. LResult ) then
         Write ( LUErr, *) 'Par3_Make_KOp:  Unable to compute ',
     $      'blocking for KOp'
         STOP
      EndIf
C
C     Determine the dimensions of the conceptual rectangle of
C     processors supporting this distribution, and figure out
C     where we are.
C
C     NB:  NProc_Row is used by the stmt. fn. WHICH_PROC to convert
C     a basis fn. address into a processor number.  DON`T DELETE
C     THIS ASSIGNMENT or the function won't work!
C
      NProc_Row = (NAO-1) / RBlkSz + 1
C
C     The rest of these guys could be turned into stmt fns, but
C     it doesn't seem like there is a need at the moment.
C
      Me_Row = Mod( Me, NProc_Row) + 1
      Me_Col = Me / NProc_Row + 1
C
C     Figure out what the corner of our panel is
C
      MyKOp_Row = (Me_Row-1) * RBlkSz + 1
      MyKOp_Col = (Me_Col-1) * CBlkSz + 1
C
C     Avoid allocating space we won't use
C
      MyRowDim = Min( RBlkSz, NAO - MyKOp_Row + 1)
      MyColDim = Min( CBlkSz, NAO - MyKOp_Col + 1)
C
C     Allocate the space for the panel
C
      Len = MyRowDim * MyColDim * (IJHi - IJLo + 1)
C
      LResult = MA_Push_Get( MT_Dbl, Len, 'Local KOp panel',
     $   MyKOp_Hdl, MyKOp_Ind)
C
      Write (LuOut, 9010) Me, MyKOp_Row, MyKOp_Row+MyRowDim-1,
     $   MyKOp_Col, MyKOp_Col+MyColDim-1, IJLo, IJHi,
     $   MA_SizeOf( MT_Dbl, Len, MT_Byte)
 9010 Format(1X, 'KOp panel on ', I4, ' is (',
     $   I4, ':', I4, ',', I4, ':', I4, ',', I4, ':', I4, '), ',
     $   I10, ' bytes.')
C
      Call DVZER( Len, Dbl_MB( MyKOp_Ind), 1)
C
C     *******************************************************
C     * Allocate memory for integral routines and "density" *
C     *******************************************************
C
      LResult = MA_Push_Get( MT_Dbl, NAO*NAO* (IJHi-IJLo+1),
     $   'Density for KOp', D_Hdl, D_Ind)
C
      LResult = Push_TwoInt()
      LResult = Get_Index_TwoInt( IEnv, Val_Ind, P_Ind, Q_Ind, R_Ind,
     $    S_Ind, MxNInt, Scr_Ind, LScratch)
      If ( .NOT. LResult ) Then
         Write (LuErr, *) ' Bad news trying to get indices.'
         STOP
      EndIf
C
      Buf_Per_Proc = 10
      Total_Buffers = Buf_Per_Proc * NProc
      Call Par3_Create_Int_Pool( MxNInt, Buf_Per_Proc)
C
      Call FFFlush(6)
C
C     These are local arrays used in processing the integrals
C
      LResult = MA_Push_Get( MT_Int, Total_Buffers,
     $   'Local ready buffer list', Ready_Hdl, Ready_Ind)
      LResult = MA_Push_Get( MT_Int, NProc, 'Processing completed list',
     $   Processed_Hdl, Processed_Ind)
C
C     ********************************************
C     * Control massaging of integrals by TwoInt *
C     ********************************************
C     The indices need not be completely canonical.
C     Better to remove zero integrals once than process them many times???
C
      Canonical_Ints = .TRUE.
      Nonzero_Ints = .TRUE.
      Perm_Sym = .TRUE.
C
C     *****************************************
C     * Loop unique shells in canonical order *
C     *****************************************
C
      Next_Batch = NxTask( NProc, Int_Batch_Granularity)
      PQRS = 0
      Do PSh = 1, NShell
         Do QSh = 1, PSh
            Do RSh = 1, PSh
C
               SShHi = RSh
               If (PSh .eq. RSh) SShHi = QSh
               Do SSh = 1, SShHi
C
                 If ( PQRS .eq. Next_Batch ) then
C
C                   See if there is a free buffer for our integrals.
C                   If not, we must process everyone else's and
C                   wait for everyone else to process ours until
C                   one becomes free.
C
                    T_Spin = T_Spin - TCGTime()
 1000               Continue
                    Free_Buffer
     $                 = Par3_Grab_Buffer(Int_MB(Processed_Ind), NProc,
     $                 Not_Free)
c$$$                    If ( Free_Buffer .eq. 0 ) then
C
                       Not_Free = Not_Free + 1
                       If (Mod( Not_Free, 100) .eq. 0) then
                          Write (0, *) 'Hey, I''ve been waiting for ',
     $                       Not_Free
                          Call FFFlush(0)
                       EndIf
C
C                      Find out what's out there for us
C
                       Call Par3_Ints_for_Me( Int_MB( Ready_Ind),
     $                    Total_Buffers, MxNInt)
C
C                      Process anything that has data for us
C
                       Do From = 1, Total_Buffers
C
                          If ( Int_MB( Ready_Ind+From-1) .gt. 0) then
                             Work = Work + 1
                             T_Spin = T_Spin + TCGTime()
C
C                            First, get the integrals
C
                             NInts = Int_MB( Ready_Ind+From-1)
                             Call Par3_Get_Ints( From, NInts,
     $                          Dbl_MB( ERI(1)),
     $                          Int_MB( Idx1(1) ), Int_MB( Idx2(1) ),
     $                          Int_MB( Idx3(1) ), Int_MB( Idx4(1) ) )
C
                             Call Par3_Process_Ints( NInts,
     $                          Dbl_MB( ERI(1)),
     $                          Int_MB( Idx1(1) ), Int_MB( Idx2(1) ),
     $                          Int_MB( Idx3(1) ), Int_MB( Idx4(1) ), C,
     $                          Dbl_MB( D(1, 1)), Dbl_MB( MyKOp_Ind),
     $                          MyKOp_Row, MyKOp_Col, MyRowDim,
     $                          MyColDim, IJLo, IJHi, NAO, RBlkSz,
     $                          CBlkSz, NProc_Row,
     $                          FPMO1, FPAO1, T_Den, T_KOp)
                             T_Spin = T_Spin - TCGTime()
                          EndIf
                       EndDo    ! Processes
C
                       Probes = Probes + 1
                       If ( Free_Buffer .eq. 0)
     $                    No_Buffers = No_Buffers + 1
C
C                      Go back and see if we have a free buffer yet
C
                       If ( Free_Buffer .eq. 0) Goto 1000
c$$$                    EndIf
                    T_Spin = T_Spin + TCGTime()
C
C                   *********************************
C                   * Generate a batch of integrals *
C                   *********************************
C
                    Not_Free = 0
C
                    T_Int = T_Int - TCGTime()
                    Call TwoInt_PermSym(IEnv, Basis,
     $                 PSh, QSh, RSh, SSh,
     $                 DBL_MB( ERI(1)),
     $                 Int_MB( Idx1(1)),
     $                 Int_MB( Idx2(1)),
     $                 Int_MB( Idx3(1)),
     $                 Int_MB( Idx4(1)),
     $                 NInts, Dbl_MB(Scr_Ind),
     $                 LScratch, Canonical_Ints, Nonzero_Ints,
     $                 Perm_Sym, FPMO1, FPAO1)
                    T_Int = T_Int + TCGTime()
C                
                    TotInt = TotInt + NInts
                    FPMO = FPMO + FPMO1
                    FPAO = FPAO + FPAO1
                    Batches_Prod = Batches_Prod + 1
                    Ints_Prod = Ints_Prod + NInts
C
C                   Post this batch of integrals so that everyone
C                   can read them.  Note that the buffer is local
C                   to this process, so there should be no
C                   communication cost here.
C                   
                    Call Par3_Put_Ints( Free_Buffer, NInts,
     $                 Dbl_MB( ERI(1)),
     $                 Int_MB( Idx1(1) ), Int_MB( Idx2(1) ),
     $                 Int_MB( Idx3(1) ), Int_MB( Idx4(1) ),
     $                 Int_MB( Processed_Ind), NProc )
C                   
C                   Find out what my next task is
C
                    Next_Batch = NxTask( NProc, Int_Batch_Granularity)
                 EndIf          ! Mod(PQRS, NProc) .eq. Me
C
C                Update counter for load (im)balancing
C
                 PQRS = PQRS + 1
              EndDo             ! SSh
           EndDo                ! RSh
           Still_Running = Par3_Check_Done(Int_MB(Ready_Ind),
     $        Total_Buffers, MxNInt)
c$$$           If ( Still_Running .lt. NProc ) then
           Write (0, '(A, I3, A, 5I5)')
     $        'I am one of ', Still_Running, ' at PQRS ',
     $           PQRS, PSh, QSh, RSh, SSh
c$$$           EndIf
        EndDo                   ! QSh
      EndDo          ! PSh
C
C     If we finish before others, they will still be looking for our
C     data, so we must tell them we're done.
C
      Call Par3_Ints_Finished( Int_MB( Ready_Ind ), Buf_per_proc,
     $   MxNInt)
C
C     *************************************************************
C     * Process everyone else's integrals until they're done too. *
C     *************************************************************
C
      Last_Pass = .FALSE.
      Last_Running = -1
      Still_Running = NProc
      T_Sync = T_Sync - TCGTime()
 3000 Continue
C
C        Find out what's out there for us
C
      Call Par3_Ints_for_Me( Int_MB( Ready_Ind), Total_Buffers, MxNInt )
C
C     Process anything that has data for us
C
      Do From = 1, Total_Buffers
C
         If ( Int_MB( Ready_Ind+From-1) .gt. 0) then
            T_Sync = T_Sync + TCGTime()
C
C           First, get the integrals
C
            NInts = Int_MB( Ready_Ind+From-1)
            Call Par3_Get_Ints( From, NInts,
     $         Dbl_MB( ERI(1)),
     $         Int_MB( Idx1(1) ), Int_MB( Idx2(1) ),
     $         Int_MB( Idx3(1) ), Int_MB( Idx4(1) ) )
C
            Call Par3_Process_Ints( NInts,
     $         Dbl_MB( ERI(1)),
     $         Int_MB( Idx1(1) ), Int_MB( Idx2(1) ),
     $         Int_MB( Idx3(1) ), Int_MB( Idx4(1) ), C,
     $         Dbl_MB( D(1, 1)), Dbl_MB( MyKOp_Ind),
     $         MyKOp_Row, MyKOp_Col, MyRowDim,
     $         MyColDim, IJLo, IJHi, NAO, RBlkSz,
     $         CBlkSz, NProc_Row,
     $         FPMO1, FPAO1, T_Den, T_KOp)
            T_Sync = T_Sync - TCGTime()
         EndIf
      EndDo                     ! Processes
C
C     See if everyone has finished yet.
C
      Last_Running = Still_Running
      Still_Running = Par3_Check_Done( Int_MB(Ready_Ind), Total_Buffers,
     $   MxNInt)
C
      If (Still_Running .ne. Last_Running) then
         Write (0, *) 'Still waiting on ', Still_running
         Call FFFlush(0)
      EndIf
C
C     If everyone's done, we must do a final pass.
C
      If ( Still_Running .eq. 0 .AND. .NOT. Last_Pass) then
         Write (0, *) 'Starting last pass...'
         Call FFFlush(0)
         Last_Pass = .TRUE.
         Goto 3000
      EndIf
C
C     If some processes are still generating integrals, we must
C     process them
C
      If (Still_Running .ne. 0) Goto 3000
      T_Sync = T_Sync + TCGTime()
C
C     Free the buffers we have been using to communicate integrals
C
      Write (0, *) 'Trying to sync with everyone...'
      Call FFFlush(0)
      Next_Batch = NxTask( -NProc, Int_Batch_Granularity)
c$$$      Call GA_Sync()
C
      LResult = MA_Pop_Stack( Processed_Hdl)
      LResult = MA_Pop_Stack( Ready_Hdl)
      LResult = Pop_TwoInt()
C
      Call Par3_Destroy_Int_Pool
C
C     Compute checksums on the KOp fragments
C
      Write (0, *) 'Computing checksums...'
      Call FFFlush(0)
      Do IJ = IJLo, IJHi
         Fact = DNRM2( MyRowDim * MyColDim,
     $      Dbl_MB( MyKOp_Ind + (IJ-1) * MyRowDim * MyColDim), 1) ** 2
         Call DGOp( 12345, Fact, 1, '+')
         Write (6, *) 'KOp checksum ', IJ, Sqrt( Fact )
         Call FFFLush(0)
         Call FFFLush(6)
      EndDo
C
C
C     *************************************************
C     * Prepare global KOp to accumulate local pieces *
C     *************************************************
C
      LResult = GA_Create( MT_Dbl, NAO*NAO, NOcc*(NOcc+1)/2,
     $   'Exchange operators', NAO*NAO, 1, KOp_Hdl)
C
      Call GA_Zero( KOp_Hdl)
C
C     Now we need to put the local KOp panels into the global space.
C     Unfortunately, the panels we're using locally don't correspond
C     to contiguous elements of the global KOp structure, so we
C     must break it down quite a lot.
C
      Write (0, *) 'Transposing...'
      Call FFFlush(0)
      T_Upd = T_Upd - TCGTime()
      Do IJ = IJLo, IJHi
         Do Q = 1, MyColDim
            Call GA_Acc( KOp_Hdl, MyKOp_Row + (Q+MyKOp_Col-2)*NAO,
     $         MyKOp_Row + (Q+MyKOp_Col-2)*NAO + MyRowDim - 1,
     $         IJ, IJ, Dbl_MB( MyKOp_Ind + (Q-1)*MyRowDim +
     $         (IJ-1)*MyRowDim*MyColDim), MyRowDim, FP_One)
         EndDo
      EndDo
      T_Upd = T_Upd + TCGTime()
C
      Write (0, *) 'Computing checksums...'
      Call FFFlush(0)
      Call GA_Sync()
      Do IJ = IJLo, IJHi
         Call GA_Get(KOp_Hdl, 1, NAO*NAO, IJ, IJ,
     $      Dbl_MB( D_Ind), NAO*NAO)
         Write (6, *) 'KOp checksum ', IJ, DNRM2( NAO*NAO,
     $      Dbl_MB( D_Ind), 1)
         Call FFFLush(0)
         Call FFFLush(6)
      EndDo
C
C     Free up dynamically allocated work space
C
      LResult = MA_Pop_Stack( D_Hdl)
C
      LResult = MA_Pop_Stack( MyKOp_Hdl )
C
C
      Write (6, *) 'T_Int =  ', T_Int
      Write (6, *) 'T_Den =  ', T_Den
      Write (6, *) 'T_KOp =  ', T_KOp
      Write (6, *) 'T_Upd =  ', T_Upd
      Write (6, *) 'T_Spin = ', T_Spin
      Write (6, *) 'T_Sync = ', T_Sync
      Write (6, *) 'Count = ', PQRS, Count
C
      Write (6, *) 'I produced ', Ints_Prod, ' integrals in ',
     $   Batches_Prod, ' batches.'
      Write (6, *) 'I made ', Probes, ' probes and found work ',
     $   Work, ' times and no buffers ', No_Buffers, ' times.'
C
C     Need to synchronize here to insure that the global KOp array is
C     complete and correct before it is used in anything.
C
      Call GA_Sync()
      Return
      End
