C:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
C NAME
C     Toy2_Eval_Resid -- Evaluate the residual of the amplitude equations
C
C SYNOPSIS
      Subroutine Toy2_Eval_Resid( NExt, NInt, KOp, T2, FInt, LDFInt,
     $   FExt, LDFExt, S, LDS, SInv, LDSInv, Scratch, R,
     $   FPMO, FPAO)
      Implicit NONE
      Integer NExt, NInt, LDFInt, LDFExt, LDS, LDSInv
      Double precision FPMO, FPAO
      Double Precision KOp(NExt, NExt, NInt*(NInt+1)/2),
     $   T2( NExt, NExt, NInt*(NInt+1)/2),
     $   FInt(LDFInt, NInt), FExt(LDFExt, NExt), S(LDS, NExt),
     $   SInv(LDSInv, NExt), Scratch(NExt, NExt),
     $   R(NExt, NExt, NInt*(NInt+1)/2)
C
C ARGUMENTS
C     NExt    Dimension of the external orbital space (IN)
C     NInt    Dimension of the internal (occupied) orbital space (IN)
C     KOp     Exchange operators (IN)
C     T2      Amplitudes to be used for evaluation of residual (IN)
C     FInt    Fock matrix in internal (occupied) basis (IN)
C     LDFInt  Leading dimension of FInt (IN)
C     FExt    Fock matrix in external (usu. AO) basis (IN)
C     LDFExt  Leading dimension of FExt (IN)
C     S       Overlap matrix in external basis (IN)
C     LDS     Leading dimension of S (IN)
C     SInv    Inverse of S (IN)
C     LDSInv  Leading dimension of SInv (IN)
C     Scratch An NExt*NExt scratch matrix (Scratch)
C     R       Residual in external basis (OUT)
C     FPMO    Number of FP multiply operations (OUT)
C     FPAO    Number of FP add operations (OUT)
C
C DESCRIPTION
C     Evaluate the "residual", r_{ij}:
C     S r_{ij} S = K_{ij} - K_{ij}^\dagger + F t_{ij} S + S t_{ij} F 
C                - S ( f_{im} t_{mj} + t_{im} f_{mj} ) S
C
C     For efficiency, this is implemented as:
C     r_{ij} = S^{-1} [ K_{ij} - K_{ij}^\dagger + F t_{ij} S 
C            + S t_{ij} F ] S^{-1} - f_{im} t_{mj} + t_{im} f_{mj}
C     which is 2 matrix multiplications per $ij$ cheaper.
C
C     This routine is roughly O( 6 NExt^3 NInt(NInt+1)/2 ) FP
C     multiplications and additions.  The actual numbers are accumulated
C     and returned in FPMO and FPAO.
C
C NOTES
C     An optimization for canonical internal orbitals, or more
C     generally for small f * t would help reduce the work in the
C     last two terms (the four loops on M, below).  The general version
C     would require some input about the size of t_{ij}, but this
C     could easily be evaluated as bart of the convergence test.
C     For storage efficiency, it could even be compressed.
C
C     An optimization for the case of orthogonal external orbitals
C     would be to avoid multiplication by S and SInv.  The problem
C     is that this would really require a different routine, since
C     the data movement would not be optimal (copy operations between
C     Scratch and R would have to be done instead of the multiplications.
C
C     In a similar vein, perhaps it would be useful for the caller to
C     pass arguments indicating if either (internal or external) Fock
C     matrices are diagonal in order to bypass some operations.
C     Once again, this starts to be a different routine pretty quickly.
C:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#include "numerical_constants.fh"
#include "precision.h"
C     
C LOCAL VARIABLES
      Integer I, J, M, IJ, IM, MJ
C
      FPMO = FP_Zero
      FPAO = FP_Zero
C
      Do I = 1, NInt
         Do J = 1, I
            IJ = I * (I-1)/2 + J
C
c           ****************************
C           * Operations for KOp terms *
c           ****************************
C
            FPAO = FPAO + 3 * NExt*NExt
C
C           ********************************************************
C           * Alpha-alpha + beta-beta integral (KOp) contributions *
C           ********************************************************
C           
            Call DGEMA('N', NExt, NExt,
     $         FP_One, KOp(1, 1, IJ), NExt,
     $         FP_Zero, R(1, 1, IJ), NExt)
            Call DGEMA('T', NExt, NExt,
     $         -FP_One, KOp(1, 1, IJ), NExt,
     $         FP_One, R(1, 1, IJ), NExt)
C           
C           ***************************
C           * Alpha-beta contribution *
C           ***************************
C
            Call DGEMA('N', NExt, NExt,
     $         FP_One, KOp(1, 1, IJ), NExt,
     $         FP_One, R(1, 1, IJ), NExt)
C
C           **************************************
C           * Operations for F t S + S t F terms *
C           **************************************
C
            FPMO = FPMO + 4 * NExt * NExt * NExt
            FPAO = FPAO + 4 * NExt * NExt * NExt
C
C           ***************************************
C           * r(ij,ab) <-- + F(ae) T(ij,ef) S(fb) *
C           *              + S(af) T(ij,fe) F(eb) *
C           ***************************************
C
            Call DGEMM( 'N', 'N', NExt, NExt, NExt, FP_One,
     $         FExt, LDFExt, T2(1, 1, IJ), NExt,
     $         FP_Zero, Scratch, NExt)
            Call DGEMM( 'N', 'N', NExt, NExt, NExt, FP_One,
     $         Scratch, NExt, S, LDS, FP_One, R(1, 1, IJ), NExt)
C
            Call DGEMM( 'N', 'N', NExt, NExt, NExt, FP_One,
     $           S, LDS, T2(1, 1, IJ), NExt, FP_Zero, Scratch, NExt)
            Call DGEMM( 'N', 'N', NExt, NExt, NExt, FP_One,
     $         Scratch, NExt, FExt, LDFExt, FP_One, R(1, 1, IJ), NExt)
C
C           **************************************
C           * Operations for back transformation *
C           **************************************
C
            FPMO = FPMO + 2 * NExt * NExt * NExt
            FPAO = FPAO + 2 * NExt * NExt * NExt
C
C           *******************************************************
C           * Transform Scr to AO basis [ S r_{ij} S --> r_{ij} ] *
C           *******************************************************
C           Do this now to avoid having to multiply the inter-pair
C           terms by S then immediately by S^-1.
C
            Call DGEMM( 'N', 'N', NExt, NExt, NExt,
     $         FP_One, SInv, LDSInv, R(1, 1, IJ), NExt,
     $         FP_Zero, Scratch, NExt)
            Call DGEMM( 'N', 'N', NExt, NExt, NExt,
     $         FP_One, Scratch, NExt, SInv, LDSInv,
     $         FP_Zero, R(1, 1, IJ), NExt)
C
C           *****************************************************
C           * Interaction of other occupied pairs with this one *
C           *****************************************************
C           Note that t_{ji} = t_{ij}^T, so we must break up
C           the range of M in order to do put each piece of t into
C           the summation in the proper orientation.
C
            Do M = 1, J-1
               MJ = Max( M, J) * (Max(M, J) - 1 )/2 + Min(M, J)
C
C              ***********************************
C              * r(ij,ab) <-- -FInt(im) T(mj,ab) *
C              ***********************************
C
               FPMO = FPMO + NExt * NExt
               FPAO = FPAO + NExt * NExt
C
               Call DGEMA('T', NExt, NExt,
     $            -FInt(I, M), T2(1, 1, MJ), NExt,
     $            FP_One, R(1, 1, IJ), NExt)
            EndDo   ! M
C
            Do M = J, NInt
               MJ = Max( M, J) * (Max(M, J) - 1 )/2 + Min(M, J)
C
C              ***********************************
C              * r(ij,ab) <-- -FInt(im) T(mj,ab) *
C              ***********************************
C
               FPMO = FPMO + NExt * NExt
               FPAO = FPAO + NExt * NExt
               Call DGEMA('N', NExt, NExt,
     $            -FInt(I, M), T2(1, 1, MJ), NExt,
     $            FP_One, R(1, 1, IJ), NExt)
            EndDo   ! M
C
            Do M = 1, I
               IM = Max( M, I) * (Max(M, I) - 1 )/2 + Min(M, I)
C
C              ***********************************
C              * r(ij,ab) <-- -T(im,ab) FInt(mj) *
C              ***********************************
C
               FPMO = FPMO + NExt * NExt
               FPAO = FPAO + NExt * NExt
               Call DGEMA('N', NExt, NExt,
     $            -FInt(M, J), T2(1, 1, IM), NExt,
     $            FP_One, R(1, 1, IJ), NExt)
            EndDo   ! M 
C
            Do M = I+1, NInt
               IM = Max( M, I) * (Max(M, I) - 1 )/2 + Min(M, I)
C
C              ***********************************
C              * r(ij,ab) <-- -T(im,ab) FInt(mj) *
C              ***********************************
C
               FPMO = FPMO + NExt * NExt
               FPAO = FPAO + NExt * NExt
               Call DGEMA('T', NExt, NExt,
     $            -FInt(M, J), T2(1, 1, IM), NExt,
     $            FP_One, R(1, 1, IJ), NExt)
            EndDo   ! M 
C
         EndDo   ! J
      EndDo      ! I
C
      Return
      End

