C:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
C NAME
C     Toy2_Next_Amp -- Construct next approximation to the amplitudes
C
C SYNOPSIS
      Subroutine Toy2_Next_Amp( NExt, NInt, Damp, Step, Q, LDQ, T2,
     $   MxNorm, FNorm, Scratch, FPMO, FPAO)
      Implicit NONE
      Integer NInt, NExt, LDQ
      Double Precision Damp, MxNorm, FNorm, FPMO, FPAO
      Double Precision Step( NExt, NExt, NInt*(NInt+1)/2),
     $   Q(LDQ, NExt), T2(NExt, NExt, NInt*(NInt+1)/2),
     $   Scratch(NExt, NExt)
C
C ARGUMENTS
C     NExt     Dimension of the external space (IN)
C     NInt     Dimension of the internal (occupied) space (IN)
C     Damp     Damping factor, T2(out) = T2(in) + Damp * Step (IN)
C     Step     Computed update (residual with denominators) (IN)
C     Q        Projection operator to eliminate contributions
C              due to internal orbtials (IN)
C     LDQ      Leading dimension of Q (IN)
C     T2       Amplitudes (INOUT)
C     MxNorm   Largest elementwise change in amplitudes (OUT)
C     FNorm    F-norm of the change in amplitudes (square root of the
C              sum of the squares of the elements) (OUT)
C     Scratch  A scratch matrix (Scratch)
C     FPMO     Number of FP multiply operations (OUT)
C     FPAO     Number of FP add operations (OUT)
C
C DESCRIPTION
C     Uses the computed update and amplitudes to compute a new set of
C     amplitudes for the next iteration of the equations.  The resulting
C     amplitudes are projected to eliminate contributions from the
C     internal space.  Also returns information about the size of the
C     step actually taken.
C
C     At present, only simple static damping is available.  Eventually
C     this routine will probably involve some kind of conjugate
C     gradient or other approach.
C     
C     Every IJ pair is independent in this routine.
C
C NOTES
C     Perhaps it would be useful to provide convergence information
C     for each IJ pair?
C
C BUGS
C     DNRM2 does fancy scaling to avoid overflow, but the operation
C     count considers only the simplest possible implementation.
C:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#include "numerical_constants.fh"
#include "precision.h"
C FUNCTIONS
      Double Precision DNRM2, DGEAMX
      External DNRM2, DGEAMX
      Integer IDAMAX
      External IDAMAX
C
C LOCAL VARIABLES
      Integer I, J, IJ, IErr
C
      FPMO = FP_Zero
      FPAO = FP_Zero
C
C     **************************
C     * Scale step for damping *
C     **************************
C
      FPMO = FPMO + NExt * NExt * NInt * (NInt+1) / 2
C
      Call DSCAL( NExt*NExt*NInt*(NInt+1)/2, Damp, Step, 1)
C
C     *************************
C     * Check for convergence *
C     *************************
C     DNRM2 uses scaling, etc., so the operation count here is likely
C     wrong.
C
      FPMO = FPMO + NExt * NExt * NInt * (NInt+1) / 2
      FPAO = FPAO + NExt * NExt * NInt * (NInt+1) / 2
C
      FNorm = DNRM2( NExt*NExt*NInt*(NInt+1)/2, Step, 1)
C
      MxNorm = FP_Zero
C
      Do I = 1, NInt
         Do J = 1, I
            IJ = I * (I-1)/2 + J
C
C           ***********************************
C           * Collect convergence information *
C           ***********************************
C           Only comparisons in this function, no FP operations.
C
            MxNorm = Max( MxNorm,
     $         DGEAMX( NExt, NExt, Step(1, 1, IJ), NExt) )
C
C           *********************
C           * Update amplitudes *
C           *********************
C
            FPAO = FPAO + NExt * NExt
C
            Call DGEMA('N', NExt, NExt,
     $         FP_One, Step(1, 1, IJ), NExt,
     $         FP_One, T2(1, 1, IJ), NExt)
C
C           ***********************************
C           * Project amplitudes to insure    *
C           * orthogonality to internal space *
C           ***********************************
C
            FPMO = FPMO + 2 * NExt * NExt * NExt
            FPAO = FPAO + 2 * NExt * NExt * NExt
C
            Call DGEMM( 'T', 'N', NExt, NExt, NExt,
     $         FP_One, Q, LDQ, T2(1, 1, IJ), NExt,
     $         FP_Zero, Scratch, NExt)
            Call DGEMM( 'N', 'N', NExt, NExt, NExt,
     $         FP_One, Scratch, NExt, Q, LDQ,
     $         FP_Zero, T2(1, 1, IJ), NExt)
C
         EndDo
      EndDo
C
      Return
      End
